# Copyright 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for `glance.wsgi`."""

import http.client as http
import os
import socket
import time

from oslo_serialization import jsonutils
from oslo_utils.fixture import uuidsentinel as uuids
import requests

from glance.common import wsgi
from glance.tests import functional


class TestWSGIServer(functional.FunctionalTest):
    """WSGI server tests."""
    def test_client_socket_timeout(self):
        self.config(workers=0)
        self.config(client_socket_timeout=1)
        """Verify connections are timed out as per 'client_socket_timeout'"""
        greetings = b'Hello, World!!!'

        def hello_world(env, start_response):
            start_response('200 OK', [('Content-Type', 'text/plain')])
            return [greetings]

        server = wsgi.Server()
        server.start(hello_world, 0)
        port = server.sock.getsockname()[1]

        def get_request(delay=0.0):
            # Socket timeouts are handled rather inconsistently on Windows.
            # recv may either return nothing OR raise a ConnectionAbortedError.
            exp_exc = OSError if os.name == 'nt' else ()

            try:
                sock = socket.socket()
                sock.connect(('127.0.0.1', port))
                time.sleep(delay)
                sock.send(b'GET / HTTP/1.1\r\nHost: localhost\r\n\r\n')
                return sock.recv(1024)
            except exp_exc:
                return None

        # Should succeed - no timeout
        self.assertIn(greetings, get_request())
        # Should fail - connection timed out so we get nothing from the server
        self.assertFalse(get_request(delay=1.1))


class StagingCleanupBase:
    def _configure_api_server(self):
        self.my_api_server.deployment_flavor = 'noauth'

    def _url(self, path):
        return 'http://127.0.0.1:%d%s' % (self.api_port, path)

    def _headers(self, custom_headers=None):
        base_headers = {
            'X-Identity-Status': 'Confirmed',
            'X-Auth-Token': '932c5c84-02ac-4fe5-a9ba-620af0e2bb96',
            'X-User-Id': 'f9a41d13-0c13-47e9-bee2-ce4e8bfe958e',
            'X-Tenant-Id': uuids.tenant1,
            'X-Roles': 'reader,member',
        }
        base_headers.update(custom_headers or {})
        return base_headers

    def test_clean_on_start(self):
        staging = os.path.join(self.test_dir, 'staging')

        # Start the server
        self.start_servers(**self.__dict__.copy())

        # Create an image
        path = self._url('/v2/images')
        headers = self._headers({'content-type': 'application/json'})
        data = jsonutils.dumps({'name': 'image-1', 'type': 'kernel',
                                'disk_format': 'aki',
                                'container_format': 'aki'})
        response = requests.post(path, headers=headers, data=data)
        self.assertEqual(http.CREATED, response.status_code)
        image = jsonutils.loads(response.text)
        image_id = image['id']

        # Stage data for the image
        path = self._url('/v2/images/%s/stage' % image_id)
        headers = self._headers({'Content-Type': 'application/octet-stream'})
        image_data = b'ZZZZZ'
        response = requests.put(path, headers=headers, data=image_data)
        self.assertEqual(http.NO_CONTENT, response.status_code)

        # Stop the server
        self.my_api_server.stop()

        # Create more files in staging, one unrecognized one, and one
        # uuid that matches nothing in the database, and some residue
        # like we would see from failed conversions and decompressions
        # for the image we created above.
        open(os.path.join(staging, 'foo'), 'w')
        open(os.path.join(staging, uuids.stale), 'w')
        open(os.path.join(staging, uuids.converting), 'w')
        converting_fn = os.path.join(staging, '%s.qcow2' % uuids.converting)
        decompressing_fn = os.path.join(staging, '%s.uc' % uuids.decompressing)
        open(converting_fn, 'w')
        open(decompressing_fn, 'w')

        # Restart the server. We set "needs_database" to False here to avoid
        # recreating the database during startup, thus causing the server to
        # think there are no valid images and deleting everything.
        self.my_api_server.needs_database = False
        self.start_with_retry(self.my_api_server,
                              'api_port', 3, **self.__dict__.copy())

        # Poll to give it time to come up and do the work. Use the presence
        # of the extra files to determine if the cleanup has run yet.
        for i in range(0, 10):
            try:
                requests.get(self._url('/v2/images'))
            except Exception:
                # Not even answering queries yet
                pass
            else:
                files = os.listdir(staging)
                if len(files) == 2:
                    break

            time.sleep(1)

        # We should still find the not-an-image file...
        self.assertTrue(os.path.exists(os.path.join(staging, 'foo')))
        # ...and make sure the actually-staged image file is still present....
        self.assertTrue(os.path.exists(os.path.join(staging, image_id)))
        # ... but the stale image should be gone,
        self.assertFalse(os.path.exists(os.path.join(staging,
                                                     uuids.stale)))
        # ... along with the residue of the conversion ...
        self.assertFalse(os.path.exists(converting_fn))
        # ... and the residue of the decompression.
        self.assertFalse(os.path.exists(decompressing_fn))

        self.stop_servers()


class TestStagingCleanupMultistore(functional.MultipleBackendFunctionalTest,
                                   StagingCleanupBase):
    """Test for staging store cleanup on API server startup.

    This tests the multistore case.
    """
    def setUp(self):
        super(TestStagingCleanupMultistore, self).setUp()
        self.my_api_server = self.api_server_multiple_backend
        self._configure_api_server()


class TestStagingCleanupSingleStore(functional.FunctionalTest,
                                    StagingCleanupBase):
    """Test for staging store cleanup on API server startup.

    This tests the single store case.
    """
    def setUp(self):
        super(TestStagingCleanupSingleStore, self).setUp()
        self.my_api_server = self.api_server
        self._configure_api_server()



# Copyright 2024 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import os

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging

from glance.common import exception
from glance import context
import glance.db
from glance.i18n import _
from glance.image_cache.drivers import common


LOG = logging.getLogger(__name__)
CONF = cfg.CONF
CONF.import_opt("image_cache_sqlite_db", "glance.image_cache.drivers.sqlite")


def can_migrate_to_central_db():
    # Return immediately if cache is disabled
    if not (CONF.paste_deploy.flavor and 'cache' in CONF.paste_deploy.flavor):
        return False

    is_centralized_db_driver = CONF.image_cache_driver == "centralized_db"
    # Check worker_self_reference_url is set if cache is enabled and
    # cache driver is centralized_db
    if is_centralized_db_driver and not CONF.worker_self_reference_url:
        msg = _("'worker_self_reference_url' needs to be set "
                "if `centralized_db` is defined as cache driver "
                "for image_cache_driver config option.")
        raise RuntimeError(msg)

    return is_centralized_db_driver


def migrate_if_required():
    if can_migrate_to_central_db():
        sqlite_db_file = get_db_path()
        if sqlite_db_file:
            LOG.info("Initiating migration process from SQLite to Centralized "
                     "database")
            migrate = Migrate(sqlite_db_file, glance.db.get_api())
            migrate.migrate()


def get_db_path():
    """Return the local path to sqlite database."""
    db = CONF.image_cache_sqlite_db
    base_dir = CONF.image_cache_dir
    db_file = os.path.join(base_dir, db)
    if not os.path.exists(db_file):
        LOG.debug('SQLite caching database not located, skipping migration')
        return

    return db_file


class Migrate:
    def __init__(self, db, db_api):
        self.db = db
        self.db_api = db_api
        self.context = context.get_admin_context()
        self.node_reference = CONF.worker_self_reference_url

    @lockutils.synchronized('sqlite_centralized_migrate', external=True)
    def migrate(self):
        LOG.debug("Adding local node reference %(node)s in centralized db",
                  {'node': self.node_reference})
        to_be_deleted = []
        try:
            self.db_api.node_reference_create(
                self.context, self.node_reference)
        except exception.Duplicate:
            LOG.debug("Node reference %(node)s is already recorded, "
                      "ignoring it", {'node': self.node_reference})

        LOG.debug("Connecting to SQLite db %s", self.db)
        with common.get_db(self.db) as sqlite_db:
            cur = sqlite_db.execute("""SELECT
                             image_id, hits, last_accessed, last_modified, size
                             FROM cached_images
                             ORDER BY image_id""")
            cur.row_factory = common.dict_factory
            for r in cur:
                # NOTE(abhishekk): Check if cache record is already present for
                # current node in centralized db
                if not self.db_api.is_image_cached_for_node(
                        self.context, self.node_reference, r['image_id']):
                    LOG.debug("Migrating image %s from SQLite to Centralized "
                              "db.", r['image_id'])
                    # NOTE(abhishekk): Converting dates to be compatible with
                    # centralized db
                    last_accessed = datetime.datetime.utcfromtimestamp(
                        r['last_accessed']).isoformat()
                    last_modified = datetime.datetime.utcfromtimestamp(
                        r['last_modified']).isoformat()
                    # insert into centralized database
                    self.db_api.insert_cache_details(
                        self.context, self.node_reference, r['image_id'],
                        r['size'], hits=r['hits'], last_modified=last_modified,
                        last_accessed=last_accessed)

                    # Verify entry is made in centralized db before adding
                    # image id to list to delete later from sqlite db
                    if self.db_api.is_image_cached_for_node(
                            self.context, self.node_reference, r['image_id']):
                        LOG.debug("Image %(uuid)s is migrated to centralized "
                                  "db for node %(node)s",
                                  {'uuid': r['image_id'],
                                   'node': self.node_reference})
                        to_be_deleted.append(r['image_id'])
                else:
                    LOG.debug('Skipping migrating image %(uuid)s from SQLite '
                              'to Centralized db for node %(node)s as it is '
                              'present in Centralized db.',
                              {'uuid': r['image_id'],
                               'node': self.node_reference})

            # Delete the images from sqlite db which are migrated to
            # centralized db
            for image_id in to_be_deleted:
                LOG.debug("Deleting image %s from SQLite db", image_id)
                sqlite_db.execute("""DELETE FROM cached_images
                                  WHERE image_id = ?""", (image_id,))
                sqlite_db.commit()

        if to_be_deleted:
            LOG.debug("Migrated %d records from SQLite db to Centralized "
                      "db", len(to_be_deleted))
        else:
            # NOTE(abhishekk): Safe to assume, no records present in SQLite db
            LOG.debug("No cache records found, skipping migration process.")



# Copyright 2024 RedHat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import hashlib

import glance_store as store
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import encodeutils
from oslo_utils import secretutils
from taskflow.patterns import linear_flow as lf
from taskflow import retry
from taskflow import task

import glance.async_.flows.api_image_import as image_import
from glance.common import exception
from glance.common import store_utils
from glance.i18n import _, _LW


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class _HashCalculationFailed(exception.GlanceException):

    def __init__(self, message):
        super(_HashCalculationFailed, self).__init__(message)


class _InvalidLocation(exception.GlanceException):

    def __init__(self, message):
        super(_InvalidLocation, self).__init__(message)


class _CalculateHash(task.Task):

    def __init__(self, task_id, task_type, image_repo, image_id,
                 hashing_algo, status=None):
        self.task_id = task_id
        self.task_type = task_type
        self.image_repo = image_repo
        self.image_id = image_id
        self.hashing_algo = hashing_algo
        self.image_status = status
        super(_CalculateHash, self).__init__(
            name='%s-CalculateHash-%s' % (task_type, task_id))

    def _calculate_hash(self, image):
        current_os_hash_value = hashlib.new(self.hashing_algo)
        current_checksum = secretutils.md5(usedforsecurity=False)
        for chunk in image.get_data():
            if chunk is None:
                break
            current_checksum.update(chunk)
            current_os_hash_value.update(chunk)
        image.checksum = current_checksum.hexdigest()
        image.os_hash_value = current_os_hash_value.hexdigest()

    def _set_checksum_and_hash(self, image):
        retries = 0
        while retries <= CONF.http_retries and image.os_hash_value is None:
            retries += 1
            try:
                self._calculate_hash(image)
                self.image_repo.save(image)
            except IOError as e:
                LOG.debug('[%i/%i] Hash calculation failed due to %s',
                          retries, CONF.http_retries,
                          encodeutils.exception_to_unicode(e))
                if retries == CONF.http_retries:
                    if image.status != 'active':
                        # NOTE(pdeore): The image location add operation
                        # should succeed so this exception should be raised
                        # only when image status is not active.
                        msg = (_('Hash calculation failed for image %s '
                                 'data') % self.image_id)
                        raise _HashCalculationFailed(msg)
                    else:
                        msg = (_LW("Hash calculation failed for image %s "
                                   "data") % self.image_id)
                        LOG.warning(msg)
            except store.exceptions.NotFound:
                # NOTE(pdeore): This can happen if image delete attempted
                # when hash calculation is in progress, which deletes the
                # image data from backend(specially rbd) but image remains
                # in 'active' state.
                # see: https://bugs.launchpad.net/glance/+bug/2045769
                # Once this ceph side issue is fixed, we'll keep only the
                # warning message here and will remove the deletion part
                # which is a temporary workaround.
                LOG.debug(_('Failed to calculate checksum of %(image_id)s '
                            'as image data has been deleted from the '
                            'backend'), {'image_id': self.image_id})
                image.delete()
                self.image_repo.remove(image)
                break

    def execute(self):
        image = self.image_repo.get(self.image_id)
        if image.status == 'queued':
            image.status = self.image_status
        image.os_hash_algo = self.hashing_algo
        self.image_repo.save(image)
        self._set_checksum_and_hash(image)

    def revert(self, result, **kwargs):
        """Set os_hash_algo to None when hash calculation fails
           and remove the location by reverting image to queued
           state
        """
        try:
            image = self.image_repo.get(self.image_id)
            if image.status == 'importing':
                if not image.locations[0]['url'].startswith("http"):
                    # NOTE(pdeore): `http` store doesn't allow deletion of
                    # location:
                    image.locations.pop()
                image.status = 'queued'
            image.os_hash_algo = None
            self.image_repo.save(image)
        except exception.NotFound:
            LOG.debug("Image %s might have been deleted from the backend",
                      self.image_id)


class _VerifyValidationData(task.Task):

    def __init__(self, task_id, task_type, image_repo, image_id,
                 val_data):
        self.task_id = task_id
        self.task_type = task_type
        self.image_repo = image_repo
        self.image_id = image_id
        self.val_data = val_data
        super(_VerifyValidationData, self).__init__(
            name='%s-VerifyValidationData-%s' % (task_type, task_id))

    def execute(self):
        """Verify the Validation Data with calculated Hash

        :param image_id: Glance Image ID
        :val_data: Validation Data provider by user
        """
        image = self.image_repo.get(self.image_id)

        if self.val_data['os_hash_value'] != image.os_hash_value:
            msg = (_("os_hash_value: (%s) not matched with actual "
                     "os_hash_value: (%s)") % (
                   self.val_data['os_hash_value'],
                   image.os_hash_value))
            raise exception.InvalidParameterValue(msg)

    def revert(self, result, **kwargs):
        """Set image status back to queued and
           set the hash values to None
        """
        try:
            image = self.image_repo.get(self.image_id)
            if not image.locations[0]['url'].startswith("http"):
                # NOTE(pdeore): `http` store doesn't allow deletion of
                # location
                image.locations.pop()
            image.status = 'queued'
            image.os_hash_algo = None
            image.os_hash_value = None
            image.checksum = None
            self.image_repo.save(image)
        except exception.NotFound:
            LOG.debug("Image %s might have been deleted from the backend",
                      self.image_id)


class _SetHashValues(task.Task):

    def __init__(self, task_id, task_type, image_repo, image_id,
                 val_data):
        self.task_id = task_id
        self.task_type = task_type
        self.image_repo = image_repo
        self.image_id = image_id
        self.val_data = val_data
        super(_SetHashValues, self).__init__(
            name='%s-SetHashValues-%s' % (task_type, task_id))

    def execute(self):
        """Set user provided hash algo and value hash properties to image
           when do_secure_hash is False.

        :param image_id: Glance Image ID
        :val_data: Validation Data provided by user
        """
        image = self.image_repo.get(self.image_id)
        for k, v in self.val_data.items():
            setattr(image, k, v)
        self.image_repo.save(image)


class _UpdateLocationTask(task.Task):

    def __init__(self, task_id, task_type, image_repo, image_id, url,
                 context):
        self.task_id = task_id
        self.task_type = task_type
        self.image_repo = image_repo
        self.image_id = image_id
        self.url = url
        self.context = context
        super(_UpdateLocationTask, self).__init__(
            name='%s-UpdateLocationTask-%s' % (task_type, task_id))

    def execute(self):
        """Update the image location

        :param image_id: Glance Image ID
        :param url: Location URL
        """
        image = self.image_repo.get(self.image_id)
        try:
            # (NOTE(pdeore): Add metadata key to add the store identifier
            # as location metadata
            updated_location = {
                'url': self.url,
                'metadata': {},
            }
            if CONF.enabled_backends:
                updated_location = store_utils.get_updated_store_location(
                    [updated_location], context=self.context)[0]

            image.locations.append(updated_location)
            self.image_repo.save(image)
        except (exception.Invalid, exception.BadStoreUri) as e:
            raise _InvalidLocation(e.msg)


class _SetImageToActiveTask(task.Task):

    def __init__(self, task_id, task_type, image_repo, image_id):
        self.task_id = task_id
        self.task_type = task_type
        self.image_repo = image_repo
        self.image_id = image_id
        super(_SetImageToActiveTask, self).__init__(
            name='%s-SetImageToActiveTask-%s' % (task_type, task_id))

    def execute(self):
        """Set Image status to Active

        :param image_id: Glance Image ID
        """
        image = self.image_repo.get(self.image_id)
        image.status = 'active'
        self.image_repo.save(image)

    def revert(self, result, **kwargs):
        """Set image status back to queued and
           remove the location if it's added.
        """
        try:
            image = self.image_repo.get(self.image_id)
            if image.status != 'active':
                if not image.locations[0]['url'].startswith("http"):
                    # NOTE(pdeore): `http` store doesn't allow deletion of
                    # location
                    image.locations.pop()
                if image.status == 'importing':
                    image.status = 'queued'
            self.image_repo.save(image)
        except exception.NotFound:
            LOG.debug("Image %s might have been deleted from the backend",
                      self.image_id)


def get_flow(**kwargs):
    """Return task flow

    :param task_id: Task ID
    :param task_type: Type of the task
    :param task_repo: Task repo
    :param image_repo: Image repository used
    :param image_id: ID of the Image to be processed
    """
    task_id = kwargs.get('task_id')
    task_type = kwargs.get('task_type')
    task_repo = kwargs.get('task_repo')
    image_repo = kwargs.get('image_repo')
    admin_repo = kwargs.get('admin_repo')
    image_id = kwargs.get('image_id')
    val_data = kwargs.get('val_data', {})
    loc_url = kwargs.get('loc_url')
    context = kwargs.get('context')

    hashing_algo = val_data.get("os_hash_algo",
                                CONF['hashing_algorithm'])

    # Instantiate an action wrapper with the admin repo if we got one,
    # otherwise with the regular repo.
    action_wrapper = image_import.ImportActionWrapper(
        admin_repo or image_repo, image_id, task_id)
    kwargs['action_wrapper'] = action_wrapper

    flow = lf.Flow(task_type, retry=retry.AlwaysRevert())
    flow.add(image_import._ImageLock(task_id, task_type, action_wrapper))
    flow.add(
        _UpdateLocationTask(task_id, task_type, image_repo, image_id,
                            loc_url, context))
    if CONF.do_secure_hash:
        if val_data:
            flow.add(
                _CalculateHash(task_id, task_type, image_repo, image_id,
                               hashing_algo, status='importing'))
            flow.add(
                _VerifyValidationData(task_id, task_type, image_repo,
                                      image_id, val_data))
            flow.add(
                _SetImageToActiveTask(task_id, task_type, image_repo,
                                      image_id))
        else:
            flow.add(
                _SetImageToActiveTask(
                    task_id, task_type, image_repo, image_id))
            flow.add(
                _CalculateHash(task_id, task_type, image_repo, image_id,
                               hashing_algo))
    elif val_data:
        flow.add(
            _SetHashValues(task_id, task_type, image_repo, image_id,
                           val_data))
        flow.add(
            _SetImageToActiveTask(task_id, task_type, image_repo, image_id))
    else:
        flow.add(
            _SetImageToActiveTask(task_id, task_type, image_repo, image_id))

    flow.add(
        image_import._CompleteTask(task_id, task_type, task_repo,
                                   action_wrapper))

    return flow



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import copy
import functools
import json
import os
import urllib.request

import glance_store as store_api
from glance_store import backend
from glance_store import exceptions as store_exceptions
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import encodeutils
from oslo_utils import excutils
from oslo_utils import timeutils
from oslo_utils import units
import taskflow
from taskflow.patterns import linear_flow as lf
from taskflow import retry
from taskflow import task

from glance.api import common as api_common
import glance.async_.flows._internal_plugins as internal_plugins
import glance.async_.flows.plugins as import_plugins
from glance.async_ import utils
from glance.common import exception
from glance.common.scripts.image_import import main as image_import
from glance.common.scripts import utils as script_utils
from glance.common import store_utils
from glance.i18n import _, _LE, _LI
from glance.quota import keystone as ks_quota


LOG = logging.getLogger(__name__)


CONF = cfg.CONF


api_import_opts = [
    cfg.ListOpt('image_import_plugins',
                item_type=cfg.types.String(quotes=True),
                bounds=True,
                sample_default='[no_op]',
                default=[],
                help=_("""
Image import plugins to be enabled for task processing.

Provide list of strings reflecting to the task Objects
that should be included to the Image Import flow. The
task objects needs to be defined in the 'glance/async/
flows/plugins/*' and may be implemented by OpenStack
Glance project team, deployer or 3rd party.

By default no plugins are enabled and to take advantage
of the plugin model the list of plugins must be set
explicitly in the glance-image-import.conf file.

The allowed values for this option is comma separated
list of object names in between ``[`` and ``]``.

Possible values:
    * no_op (only logs debug level message that the
      plugin has been executed)
    * Any provided Task object name to be included
      in to the flow.
""")),
]

CONF.register_opts(api_import_opts, group='image_import_opts')

glance_download_opts = [
    cfg.ListOpt('extra_properties',
                item_type=cfg.types.String(quotes=True),
                bounds=True,
                default=[
                    'hw_', 'trait:', 'os_distro', 'os_secure_boot',
                    'os_type'],
                help=_("""
Specify metadata prefix to be set on the target image when using
glance-download. All other properties coming from the source image won't be set
on the target image. If specified metadata does not exist on the source image
it won't be set on the target image. Note you can't set the os_glance prefix
as it is reserved by glance, so the related properties won't be set on the
target image.

Possible values:
    * List containing extra_properties prefixes: ['os_', 'architecture']

""")),
]

CONF.register_opts(glance_download_opts, group='glance_download_properties')


# TODO(jokke): We should refactor the task implementations so that we do not
# need to duplicate what we have already for example in base_import.py.


class _NoStoresSucceeded(exception.GlanceException):

    def __init__(self, message):
        super(_NoStoresSucceeded, self).__init__(message)


class _InvalidGlanceDownloadImageStatus(exception.GlanceException):

    def __init__(self, message):
        super(_InvalidGlanceDownloadImageStatus, self).__init__(message)


class ImportActionWrapper(object):
    """Wrapper for all the image metadata operations we do during an import.

    This is used to consolidate the changes we make to image metadata during
    an import operation, and can be used with an admin-capable repo to
    enable non-owner controlled modification of that data if desired.

    Use this as a context manager to make multiple changes followed by
    a save of the image in one operation. An _ImportActions object is
    yielded from the context manager, which defines the available operations.

    :param image_repo: The ImageRepo we should use to fetch/save the image
    :param image-id: The ID of the image we should be altering
    """

    def __init__(self, image_repo, image_id, task_id):
        self._image_repo = image_repo
        self._image_id = image_id
        self._task_id = task_id

    def __enter__(self):
        self._image = self._image_repo.get(self._image_id)
        self._image_previous_status = self._image.status
        self._assert_task_lock(self._image)

        return _ImportActions(self._image)

    def __exit__(self, type, value, traceback):
        if type is not None:
            # NOTE(danms): Do not save the image if we raised in context
            return

        # NOTE(danms): If we were in the middle of a long-running
        # set_data() where someone else stole our lock, we may race
        # with them to update image locations and erase one that
        # someone else is working on. Checking the task lock here
        # again is not perfect exclusion, but in lieu of actual
        # thread-safe location updating, this at least reduces the
        # likelihood of that happening.
        self.assert_task_lock()

        if self._image_previous_status != self._image.status:
            LOG.debug('Image %(image_id)s status changing from '
                      '%(old_status)s to %(new_status)s',
                      {'image_id': self._image_id,
                       'old_status': self._image_previous_status,
                       'new_status': self._image.status})
        self._image_repo.save(self._image, self._image_previous_status)

    @property
    def image_id(self):
        return self._image_id

    def drop_lock_for_task(self):
        """Delete the import lock for our task.

        This is an atomic operation and thus does not require a context
        for the image save. Note that after calling this method, no
        further actions will be allowed on the image.

        :raises: NotFound if the image was not locked by the expected task.
        """
        image = self._image_repo.get(self._image_id)
        self._image_repo.delete_property_atomic(image,
                                                'os_glance_import_task',
                                                self._task_id)

    def _assert_task_lock(self, image):
        task_lock = image.extra_properties.get('os_glance_import_task')
        if task_lock != self._task_id:
            LOG.error('Image %(image)s import task %(task)s attempted to '
                      'take action on image, but other task %(other)s holds '
                      'the lock; Aborting.',
                      {'image': self._image_id,
                       'task': self._task_id,
                       'other': task_lock})
            raise exception.TaskAbortedError()

    def assert_task_lock(self):
        """Assert that we own the task lock on the image.

        :raises: TaskAbortedError if we do not
        """
        image = self._image_repo.get(self._image_id)
        self._assert_task_lock(image)


class _ImportActions(object):
    """Actions available for being performed on an image during import.

    This defines the available actions that can be performed on an image
    during import, which may be done with an image owned by another user.

    Do not instantiate this object directly, get it from ImportActionWrapper.
    """

    IMPORTING_STORES_KEY = 'os_glance_importing_to_stores'
    IMPORT_FAILED_KEY = 'os_glance_failed_import'

    def __init__(self, image):
        self._image = image

    @property
    def image_id(self):
        return self._image.image_id

    @property
    def image_size(self):
        return self._image.size

    @property
    def image_locations(self):
        # Return a copy of this complex structure to make sure we do
        # not allow the plugin to mutate this underneath us for our
        # later save.  If this needs to be a thing in the future, we
        # should have moderated access like all the other things here.
        return copy.deepcopy(self._image.locations)

    @property
    def image_disk_format(self):
        return self._image.disk_format

    @property
    def image_container_format(self):
        return self._image.container_format

    @property
    def image_extra_properties(self):
        return dict(self._image.extra_properties)

    @property
    def image_status(self):
        return self._image.status

    def merge_store_list(self, list_key, stores, subtract=False):
        stores = set([store for store in stores if store])
        existing = set(
            self._image.extra_properties.get(list_key, '').split(','))

        if subtract:
            if stores - existing:
                LOG.debug('Stores %(stores)s not in %(key)s for '
                          'image %(image_id)s',
                          {'stores': ','.join(sorted(stores - existing)),
                           'key': list_key,
                           'image_id': self.image_id})
            merged_stores = existing - stores
        else:
            merged_stores = existing | stores

        stores_list = ','.join(sorted((store for store in
                                       merged_stores if store)))
        self._image.extra_properties[list_key] = stores_list
        LOG.debug('Image %(image_id)s %(key)s=%(stores)s',
                  {'image_id': self.image_id,
                   'key': list_key,
                   'stores': stores_list})

    def add_importing_stores(self, stores):
        """Add a list of stores to the importing list.

        Add stores to os_glance_importing_to_stores

        :param stores: A list of store names
        """
        self.merge_store_list(self.IMPORTING_STORES_KEY, stores)

    def remove_importing_stores(self, stores):
        """Remove a list of stores from the importing list.

        Remove stores from os_glance_importing_to_stores

        :param stores: A list of store names
        """
        self.merge_store_list(self.IMPORTING_STORES_KEY, stores, subtract=True)

    def add_failed_stores(self, stores):
        """Add a list of stores to the failed list.

        Add stores to os_glance_failed_import

        :param stores: A list of store names
        """
        self.merge_store_list(self.IMPORT_FAILED_KEY, stores)

    def remove_failed_stores(self, stores):
        """Remove a list of stores from the failed list.

        Remove stores from os_glance_failed_import

        :param stores: A list of store names
        """
        self.merge_store_list(self.IMPORT_FAILED_KEY, stores, subtract=True)

    def set_image_data(self, uri, task_id, backend, set_active,
                       callback=None):
        """Populate image with data on a specific backend.

        This is used during an image import operation to populate the data
        in a given store for the image. If this object wraps an admin-capable
        image_repo, then this will be done with admin credentials on behalf
        of a user already determined to be able to perform this operation
        (such as a copy-image import of an existing image owned by another
        user).

        :param uri: Source URL for image data
        :param task_id: The task responsible for this operation
        :param backend: The backend store to target the data
        :param set_active: Whether or not to set the image to 'active'
                           state after the operation completes
        :param callback: A callback function with signature:
                         fn(action, chunk_bytes, total_bytes)
                         which should be called while processing the image
                         approximately every minute.
        """
        if callback:
            callback = functools.partial(callback, self)
        return image_import.set_image_data(self._image, uri, task_id,
                                           backend=backend,
                                           set_active=set_active,
                                           callback=callback)

    def set_image_attribute(self, **attrs):
        """Set an image attribute.

        This allows setting various image attributes which will be saved
        upon exiting the ImportActionWrapper context.

        :param attrs: kwarg list of attributes to set on the image
        :raises: AttributeError if an attribute outside the set of allowed
                 ones is present in attrs.
        """
        allowed = ['status', 'disk_format', 'container_format',
                   'virtual_size', 'size']
        for attr, value in attrs.items():
            if attr not in allowed:
                raise AttributeError('Setting %s is not allowed' % attr)
            setattr(self._image, attr, value)

    def set_image_extra_properties(self, properties):
        """Merge values into image extra_properties.

        This allows a plugin to set additional properties on the image,
        as long as those are outside the reserved namespace. Any keys
        in the internal namespace will be dropped (and logged).

        :param properties: A dict of properties to be merged in
        """
        for key, value in properties.items():
            if key.startswith(api_common.GLANCE_RESERVED_NS):
                LOG.warning(('Dropping %(key)s=%(val)s during metadata '
                             'injection for %(image)s'),
                            {'key': key, 'val': value,
                             'image': self.image_id})
            else:
                self._image.extra_properties[key] = value

    def remove_location_for_store(self, backend):
        """Remove a location from an image given a backend store.

        Given a backend store, remove the corresponding location from the
        image's set of locations. If the last location is removed, remove
        the image checksum, hash information, and size.

        :param backend: The backend store to remove from the image
        """

        for i, location in enumerate(self._image.locations):
            if location.get('metadata', {}).get('store') == backend:
                try:
                    self._image.locations.pop(i)
                except (store_exceptions.NotFound,
                        store_exceptions.Forbidden):
                    msg = (_("Error deleting from store %(store)s when "
                             "reverting.") % {'store': backend})
                    LOG.warning(msg)
                # NOTE(yebinama): Some store drivers doesn't document which
                # exceptions they throw.
                except Exception:
                    msg = (_("Unexpected exception when deleting from store "
                             "%(store)s.") % {'store': backend})
                    LOG.warning(msg)
                else:
                    if len(self._image.locations) == 0:
                        self._image.checksum = None
                        self._image.os_hash_algo = None
                        self._image.os_hash_value = None
                        self._image.size = None
                break

    def pop_extra_property(self, name):
        """Delete the named extra_properties value, if present.

        If the image.extra_properties dict contains the named key,
        delete it.
        :param name: The key to delete.
        """
        self._image.extra_properties.pop(name, None)


class _DeleteFromFS(task.Task):

    def __init__(self, task_id, task_type):
        self.task_id = task_id
        self.task_type = task_type
        super(_DeleteFromFS, self).__init__(
            name='%s-DeleteFromFS-%s' % (task_type, task_id))

    def execute(self, file_path):
        """Remove file from the backend

        :param file_path: path to the file being deleted
        """
        if CONF.enabled_backends:
            try:
                store_api.delete(file_path, 'os_glance_staging_store')
            except store_api.exceptions.NotFound as e:
                LOG.error(_("After upload to backend, deletion of staged "
                            "image data from %(fn)s has failed because "
                            "%(em)s"), {'fn': file_path,
                                        'em': e.message})
        else:
            # TODO(abhishekk): After removal of backend module from
            # glance_store need to change this to use multi_backend
            # module.
            file_path = file_path[7:]
            if os.path.exists(file_path):
                try:
                    LOG.debug(_("After upload to the backend, deleting staged "
                                "image data from %(fn)s"), {'fn': file_path})
                    os.unlink(file_path)
                except OSError as e:
                    LOG.error(_("After upload to backend, deletion of staged "
                                "image data from %(fn)s has failed because "
                                "[Errno %(en)d]"), {'fn': file_path,
                                                    'en': e.errno})
            else:
                LOG.warning(_("After upload to backend, deletion of staged "
                              "image data has failed because "
                              "it cannot be found at %(fn)s"), {
                    'fn': file_path})


class _ImageLock(task.Task):
    def __init__(self, task_id, task_type, action_wrapper):
        self.task_id = task_id
        self.task_type = task_type
        self.action_wrapper = action_wrapper
        super(_ImageLock, self).__init__(
            name='%s-ImageLock-%s' % (task_type, task_id))

    def execute(self):
        self.action_wrapper.assert_task_lock()
        LOG.debug('Image %(image)s import task %(task)s lock confirmed',
                  {'image': self.action_wrapper.image_id,
                   'task': self.task_id})

    def revert(self, result, **kwargs):
        """Drop our claim on the image.

        If we have failed, we need to drop our import_task lock on the image
        so that something else can have a try. Note that we may have been
        preempted so we should only drop *our* lock.
        """
        try:
            self.action_wrapper.drop_lock_for_task()
        except exception.NotFound:
            LOG.warning('Image %(image)s import task %(task)s lost its '
                        'lock during execution!',
                        {'image': self.action_wrapper.image_id,
                         'task': self.task_id})
        else:
            LOG.debug('Image %(image)s import task %(task)s dropped '
                      'its lock after failure',
                      {'image': self.action_wrapper.image_id,
                       'task': self.task_id})


class _VerifyStaging(task.Task):

    # NOTE(jokke): This could be also for example "staging_path" but to
    # keep this compatible with other flows  we want to stay consistent
    # with base_import
    default_provides = 'file_path'

    def __init__(self, task_id, task_type, task_repo, uri):
        self.task_id = task_id
        self.task_type = task_type
        self.task_repo = task_repo
        self.uri = uri
        super(_VerifyStaging, self).__init__(
            name='%s-ConfigureStaging-%s' % (task_type, task_id))

        # NOTE(jokke): If we want to use other than 'file' store in the
        # future, this is one thing that needs to change.
        try:
            uri.index('file:///', 0)
        except ValueError:
            msg = (_("%(task_id)s of %(task_type)s not configured "
                     "properly. Value of node_staging_uri must be "
                     " in format 'file://<absolute-path>'") %
                   {'task_id': self.task_id,
                    'task_type': self.task_type})
            raise exception.BadTaskConfiguration(msg)

        if not CONF.enabled_backends:
            # NOTE(jokke): We really don't need the store for anything but
            # verifying that we actually can build the store will allow us to
            # fail the flow early with clear message why that happens.
            self._build_store()

    def _build_store(self):
        # TODO(abhishekk): After removal of backend module from glance_store
        # need to change this to use multi_backend module.
        # NOTE(jokke): If we want to use some other store for staging, we can
        # implement the logic more general here. For now this should do.
        # NOTE(flaper87): Due to the nice glance_store api (#sarcasm), we're
        # forced to build our own config object, register the required options
        # (and by required I mean *ALL* of them, even the ones we don't want),
        # and create our own store instance by calling a private function.
        # This is certainly unfortunate but it's the best we can do until the
        # glance_store refactor is done. A good thing is that glance_store is
        # under our team's management and it gates on Glance so changes to
        # this API will (should?) break task's tests.
        conf = cfg.ConfigOpts()
        try:
            backend.register_opts(conf)
        except cfg.DuplicateOptError:
            pass
        conf.set_override('filesystem_store_datadir',
                          CONF.node_staging_uri[7:],
                          group='glance_store')

        # NOTE(flaper87): Do not even try to judge me for this... :(
        # With the glance_store refactor, this code will change, until
        # that happens, we don't have a better option and this is the
        # least worst one, IMHO.
        store = backend._load_store(conf, 'file')

        try:
            store.configure()
        except AttributeError:
            msg = (_("%(task_id)s of %(task_type)s not configured "
                     "properly. Could not load the filesystem store") %
                   {'task_id': self.task_id, 'task_type': self.task_type})
            raise exception.BadTaskConfiguration(msg)

    def execute(self):
        """Test the backend store and return the 'file_path'"""
        return self.uri


class _ImportToStore(task.Task):

    def __init__(self, task_id, task_type, task_repo, action_wrapper, uri,
                 backend, all_stores_must_succeed, set_active):
        self.task_id = task_id
        self.task_type = task_type
        self.task_repo = task_repo
        self.action_wrapper = action_wrapper
        self.uri = uri
        self.backend = backend
        self.all_stores_must_succeed = all_stores_must_succeed
        self.set_active = set_active
        self.last_status = 0
        super(_ImportToStore, self).__init__(
            name='%s-ImportToStore-%s' % (task_type, task_id))

    def execute(self, file_path=None):
        """Bringing the imported image to back end store

        :param file_path: path to the image file
        """
        # NOTE(flaper87): Let's dance... and fall
        #
        # Unfortunately, because of the way our domain layers work and
        # the checks done in the FS store, we can't simply rename the file
        # and set the location. To do that, we'd have to duplicate the logic
        # of every and each of the domain factories (quota, location, etc)
        # and we'd also need to hack the FS store to prevent it from raising
        # a "duplication path" error. I'd rather have this task copying the
        # image bits one more time than duplicating all that logic.
        #
        # Since I don't think this should be the definitive solution, I'm
        # leaving the code below as a reference for what should happen here
        # once the FS store and domain code will be able to handle this case.
        #
        # if file_path is None:
        #    image_import.set_image_data(image, self.uri, None)
        #    return

        # NOTE(flaper87): Don't assume the image was stored in the
        # work_dir. Think in the case this path was provided by another task.
        # Also, lets try to neither assume things nor create "logic"
        # dependencies between this task and `_ImportToFS`
        #
        # base_path = os.path.dirname(file_path.split("file://")[-1])

        # NOTE(flaper87): Hopefully just scenarios #3 and #4. I say
        # hopefully because nothing prevents the user to use the same
        # FS store path as a work dir
        #
        # image_path = os.path.join(base_path, image_id)
        #
        # if (base_path == CONF.glance_store.filesystem_store_datadir or
        #      base_path in CONF.glance_store.filesystem_store_datadirs):
        #     os.rename(file_path, image_path)
        #
        # image_import.set_image_data(image, image_path, None)

        # NOTE(jokke): The different options here are kind of pointless as we
        # will need the file path anyways for our delete workflow for now.
        # For future proofing keeping this as is.

        with self.action_wrapper as action:
            self._execute(action, file_path)

    def _execute(self, action, file_path):
        self.last_status = timeutils.now()

        if action.image_status == "deleted":
            raise exception.ImportTaskError("Image has been deleted, aborting"
                                            " import.")
        try:
            action.set_image_data(file_path or self.uri,
                                  self.task_id, backend=self.backend,
                                  set_active=self.set_active,
                                  callback=self._status_callback)
        # NOTE(yebinama): set_image_data catches Exception and raises from
        # them. Can't be more specific on exceptions caught.
        except Exception:
            if self.all_stores_must_succeed:
                raise
            msg = (_("%(task_id)s of %(task_type)s failed but since "
                     "all_stores_must_succeed is set to false, continue.") %
                   {'task_id': self.task_id, 'task_type': self.task_type})
            LOG.warning(msg)
            if self.backend is not None:
                action.add_failed_stores([self.backend])

        if self.backend is not None:
            action.remove_importing_stores([self.backend])

    def _status_callback(self, action, chunk_bytes, total_bytes):
        # NOTE(danms): Only log status every five minutes
        if timeutils.now() - self.last_status > 300:
            LOG.debug('Image import %(image_id)s copied %(copied)i MiB',
                      {'image_id': action.image_id,
                       'copied': total_bytes // units.Mi})
            self.last_status = timeutils.now()

        task = script_utils.get_task(self.task_repo, self.task_id)
        if task is None:
            LOG.error(
                'Status callback for task %(task)s found no task object!',
                {'task': self.task_id})
            raise exception.TaskNotFound(self.task_id)
        if task.status != 'processing':
            LOG.error('Task %(task)s expected "processing" status, '
                      'but found "%(status)s"; aborting.')
            raise exception.TaskAbortedError()

        task.message = _('Copied %i MiB') % (total_bytes // units.Mi)
        self.task_repo.save(task)

    def revert(self, result, **kwargs):
        """
        Remove location from image in case of failure

        :param result: taskflow result object
        """
        with self.action_wrapper as action:
            action.remove_location_for_store(self.backend)
            action.remove_importing_stores([self.backend])
            if isinstance(result, taskflow.types.failure.Failure):
                # We are the store that failed, so add us to the failed list
                action.add_failed_stores([self.backend])


class _VerifyImageState(task.Task):

    def __init__(self, task_id, task_type, action_wrapper, import_method):
        self.task_id = task_id
        self.task_type = task_type
        self.action_wrapper = action_wrapper
        self.import_method = import_method
        super(_VerifyImageState, self).__init__(
            name='%s-VerifyImageState-%s' % (task_type, task_id))

    def execute(self):
        """Verify we have active image

        """
        with self.action_wrapper as action:
            if action.image_status != 'active':
                raise _NoStoresSucceeded(_('None of the uploads finished!'))

    def revert(self, result, **kwargs):
        """Set back to queued if this wasn't copy-image job."""
        with self.action_wrapper as action:
            if self.import_method != 'copy-image':
                action.set_image_attribute(status='queued')


class _CompleteTask(task.Task):

    def __init__(self, task_id, task_type, task_repo, action_wrapper):
        self.task_id = task_id
        self.task_type = task_type
        self.task_repo = task_repo
        self.action_wrapper = action_wrapper
        super(_CompleteTask, self).__init__(
            name='%s-CompleteTask-%s' % (task_type, task_id))

    def _finish_task(self, task):
        try:
            task.succeed({'image_id': self.action_wrapper.image_id})
        except Exception as e:
            # Note: The message string contains Error in it to indicate
            # in the task.message that it's a error message for the user.

            # TODO(nikhil): need to bring back save_and_reraise_exception when
            # necessary
            log_msg = _LE("Task ID %(task_id)s failed. Error: %(exc_type)s: "
                          "%(e)s")
            LOG.exception(log_msg, {'exc_type': str(type(e)),
                                    'e': encodeutils.exception_to_unicode(e),
                                    'task_id': task.task_id})

            err_msg = _("Error: %(exc_type)s: %(e)s")
            task.fail(err_msg % {'exc_type': str(type(e)),
                                 'e': encodeutils.exception_to_unicode(e)})
        finally:
            self.task_repo.save(task)

    def _drop_lock(self):
        try:
            self.action_wrapper.drop_lock_for_task()
        except exception.NotFound:
            # NOTE(danms): This would be really bad, but there is probably
            # not much point in reverting all the way back if we got this
            # far. Log the carnage for forensics.
            LOG.error('Image %(image)s import task %(task)s did not hold the '
                      'lock upon completion!',
                      {'image': self.action_wrapper.image_id,
                       'task': self.task_id})

    def execute(self):
        """Finishing the task flow

        """
        task = script_utils.get_task(self.task_repo, self.task_id)
        if task is not None:
            self._finish_task(task)
        self._drop_lock()

        LOG.info(_LI("%(task_id)s of %(task_type)s completed"),
                 {'task_id': self.task_id, 'task_type': self.task_type})


class _ImportMetadata(task.Task):

    default_provides = 'image_size'

    def __init__(self, task_id, task_type, context, action_wrapper,
                 import_req):
        self.task_id = task_id
        self.task_type = task_type
        self.context = context
        self.action_wrapper = action_wrapper
        self.import_req = import_req
        self.props_to_copy = CONF.glance_download_properties.extra_properties
        # We store the properties that will be set in case we are reverting
        self.properties = {}
        self.old_properties = {}
        self.old_attributes = {}
        super(_ImportMetadata, self).__init__(
            name='%s-ImportMetdata-%s' % (task_type, task_id))

    def execute(self):
        try:
            glance_endpoint = utils.get_glance_endpoint(
                self.context,
                self.import_req['method']['glance_region'],
                self.import_req['method']['glance_service_interface'])
            glance_image_id = self.import_req['method']['glance_image_id']
            image_download_metadata_url = '%s/v2/images/%s' % (
                glance_endpoint, glance_image_id)
            LOG.info(_LI("Fetching glance image metadata from remote host %s"),
                     image_download_metadata_url)
            token = self.context.auth_token
            request = urllib.request.Request(image_download_metadata_url,
                                             headers={'X-Auth-Token': token})
            opener = urllib.request.build_opener(
                script_utils.SafeRedirectHandler)
            with opener.open(request) as payload:
                data = json.loads(payload.read().decode('utf-8'))

            if data.get('status') != 'active':
                raise _InvalidGlanceDownloadImageStatus(
                    _('Source image status should be active instead of %s')
                    % data['status'])

            for key, value in data.items():
                for metadata in self.props_to_copy:
                    if key.startswith(metadata):
                        self.properties[key] = value

            with self.action_wrapper as action:
                # Save the old properties in case we need to revert
                self.old_properties = action.image_extra_properties
                self.old_attributes = {
                    'container_format': action.image_container_format,
                    'disk_format': action.image_disk_format,
                }

                # Set disk_format and container_format attributes
                action.set_image_attribute(
                    disk_format=data['disk_format'],
                    container_format=data['container_format'])

                # Set extra propoerties
                if self.properties:
                    action.set_image_extra_properties(self.properties)
            try:
                return int(data['size'])
            except (ValueError, KeyError):
                raise exception.ImportTaskError(
                    _('Size attribute of remote image %s could not be '
                      'determined.' % glance_image_id))
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error(
                    "Task %(task_id)s failed with exception %(error)s", {
                        "error": encodeutils.exception_to_unicode(e),
                        "task_id": self.task_id
                    })

    def revert(self, result, **kwargs):
        """Revert the extra properties set and set the image in queued"""
        with self.action_wrapper as action:
            for image_property in self.properties:
                if image_property not in self.old_properties:
                    action.pop_extra_property(image_property)
            action.set_image_extra_properties(self.old_properties)
            action.set_image_attribute(status='queued',
                                       **self.old_attributes)


def assert_quota(context, task_repo, task_id, stores,
                 action_wrapper, enforce_quota_fn,
                 **enforce_kwargs):
    try:
        enforce_quota_fn(context, context.owner, **enforce_kwargs)
    except exception.LimitExceeded as e:
        with excutils.save_and_reraise_exception():
            with action_wrapper as action:
                action.remove_importing_stores(stores)
                if action.image_status == 'importing':
                    action.set_image_attribute(status='queued')
            action_wrapper.drop_lock_for_task()
            task = script_utils.get_task(task_repo, task_id)
            if task is None:
                LOG.error(_LE('Failed to find task %r to update after '
                              'quota failure'), task_id)
            else:
                task.fail(str(e))
                task_repo.save(task)


def get_flow(**kwargs):
    """Return task flow

    :param task_id: Task ID
    :param task_type: Type of the task
    :param task_repo: Task repo
    :param image_repo: Image repository used
    :param image_id: ID of the Image to be processed
    :param uri: uri for the image file
    """
    task_id = kwargs.get('task_id')
    task_type = kwargs.get('task_type')
    task_repo = kwargs.get('task_repo')
    image_repo = kwargs.get('image_repo')
    admin_repo = kwargs.get('admin_repo')
    image_id = kwargs.get('image_id')
    import_req = kwargs.get('import_req')
    import_method = import_req['method']['name']
    uri = import_req['method'].get('uri')
    stores = kwargs.get('backend', [None])
    all_stores_must_succeed = import_req.get(
        'all_stores_must_succeed', True)
    context = kwargs.get('context')

    separator = ''
    if not CONF.enabled_backends and not CONF.node_staging_uri.endswith('/'):
        separator = '/'

    # Instantiate an action wrapper with the admin repo if we got one,
    # otherwise with the regular repo.
    action_wrapper = ImportActionWrapper(admin_repo or image_repo, image_id,
                                         task_id)
    kwargs['action_wrapper'] = action_wrapper

    if not uri and import_method in ['glance-direct', 'copy-image']:
        if CONF.enabled_backends:
            separator, staging_dir = store_utils.get_dir_separator()
            uri = separator.join((staging_dir, str(image_id)))
        else:
            uri = separator.join((CONF.node_staging_uri, str(image_id)))

    flow = lf.Flow(task_type, retry=retry.AlwaysRevert())

    flow.add(_ImageLock(task_id, task_type, action_wrapper))

    if import_method in ['web-download', 'copy-image', 'glance-download']:
        if import_method == 'glance-download':
            flow.add(_ImportMetadata(task_id, task_type,
                                     context, action_wrapper, import_req))
        internal_plugin = internal_plugins.get_import_plugin(**kwargs)
        flow.add(internal_plugin)
        if CONF.enabled_backends:
            separator, staging_dir = store_utils.get_dir_separator()
            file_uri = separator.join((staging_dir, str(image_id)))
        else:
            file_uri = separator.join((CONF.node_staging_uri, str(image_id)))
    else:
        file_uri = uri

    flow.add(_VerifyStaging(task_id, task_type, task_repo, file_uri))

    # Note(jokke): The plugins were designed to act on the image data or
    # metadata during the import process before the image goes active. It
    # does not make sense to try to execute them during 'copy-image'.
    if import_method != 'copy-image':
        for plugin in import_plugins.get_import_plugins(**kwargs):
            flow.add(plugin)
    else:
        LOG.debug("Skipping plugins on 'copy-image' job.")

    for idx, store in enumerate(stores, 1):
        set_active = (not all_stores_must_succeed) or (idx == len(stores))
        if import_method == 'copy-image':
            set_active = False
        task_name = task_type + "-" + (store or "")
        import_task = lf.Flow(task_name)
        import_to_store = _ImportToStore(task_id,
                                         task_name,
                                         task_repo,
                                         action_wrapper,
                                         file_uri,
                                         store,
                                         all_stores_must_succeed,
                                         set_active)
        import_task.add(import_to_store)
        flow.add(import_task)

    delete_task = lf.Flow(task_type).add(_DeleteFromFS(task_id, task_type))
    flow.add(delete_task)

    verify_task = _VerifyImageState(task_id,
                                    task_type,
                                    action_wrapper,
                                    import_method)
    flow.add(verify_task)

    complete_task = _CompleteTask(task_id,
                                  task_type,
                                  task_repo,
                                  action_wrapper)
    flow.add(complete_task)

    with action_wrapper as action:
        if import_method != 'copy-image':
            action.set_image_attribute(status='importing')
        image_size = (action.image_size or 0) // units.Mi
        action.add_importing_stores(stores)
        action.remove_failed_stores(stores)
        action.pop_extra_property('os_glance_stage_host')

    # After we have marked the image as intended, check quota to make
    # sure we are not over a limit, otherwise we roll back.
    if import_method == 'glance-direct':
        # We know the size of the image in staging, so we can check
        # against available image_size_total quota.
        assert_quota(kwargs['context'], task_repo, task_id,
                     stores, action_wrapper,
                     ks_quota.enforce_image_size_total,
                     delta=image_size)
    elif import_method in ('copy-image', 'web-download', 'glance-download'):
        # The copy-image, web-download and glance-download methods will use
        # staging space to do their work, so check that quota.
        assert_quota(kwargs['context'], task_repo, task_id,
                     stores, action_wrapper,
                     ks_quota.enforce_image_staging_total,
                     delta=image_size)
        assert_quota(kwargs['context'], task_repo, task_id,
                     stores, action_wrapper,
                     ks_quota.enforce_image_count_uploading)

    return flow



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
import urllib

import glance.common.exception as exception
from glance.common.scripts.image_import import main as image_import_script
from glance.common.scripts import utils
from glance.common import store_utils

import glance.tests.utils as test_utils


class TestImageImport(test_utils.BaseTestCase):

    def setUp(self):
        super(TestImageImport, self).setUp()

    def test_run(self):
        with mock.patch.object(image_import_script,
                               '_execute') as mock_execute:
            task_id = mock.ANY
            context = mock.ANY
            task_repo = mock.ANY
            image_repo = mock.ANY
            image_factory = mock.ANY
            image_import_script.run(task_id, context, task_repo, image_repo,
                                    image_factory)

        mock_execute.assert_called_once_with(task_id, task_repo, image_repo,
                                             image_factory)

    def test_import_image(self):
        image_id = mock.ANY
        image = mock.Mock(image_id=image_id)
        image_repo = mock.Mock()
        image_repo.get.return_value = image
        image_factory = mock.ANY
        task_input = mock.Mock(image_properties=mock.ANY)
        uri = mock.ANY
        with mock.patch.object(image_import_script,
                               'create_image') as mock_create_image:
            with mock.patch.object(image_import_script,
                                   'set_image_data') as mock_set_img_data:
                mock_create_image.return_value = image
                self.assertEqual(
                    image_id,
                    image_import_script.import_image(image_repo, image_factory,
                                                     task_input, None, uri))
                # Check image is in saving state before image_repo.save called
                self.assertEqual('saving', image.status)
                self.assertTrue(image_repo.save.called)
                mock_set_img_data.assert_called_once_with(image, uri, None)
                self.assertTrue(image_repo.get.called)
                self.assertTrue(image_repo.save.called)

    def test_create_image(self):
        image = mock.ANY
        image_repo = mock.Mock()
        image_factory = mock.Mock()
        image_factory.new_image.return_value = image

        # Note: include some base properties to ensure no error while
        # attempting to verify them
        image_properties = {'disk_format': 'foo',
                            'id': 'bar'}

        self.assertEqual(image,
                         image_import_script.create_image(image_repo,
                                                          image_factory,
                                                          image_properties,
                                                          None))

    @mock.patch.object(utils, 'get_image_data_iter')
    def test_set_image_data_http(self, mock_image_iter):
        uri = 'http://www.example.com'
        image = mock.Mock()
        mock_image_iter.return_value = test_utils.FakeHTTPResponse()
        self.assertIsNone(image_import_script.set_image_data(image,
                                                             uri,
                                                             None))

    def test_set_image_data_http_error(self):
        uri = 'blahhttp://www.example.com'
        image = mock.Mock()
        self.assertRaises(urllib.error.URLError,
                          image_import_script.set_image_data, image, uri, None)

    @mock.patch.object(image_import_script, 'create_image')
    @mock.patch.object(image_import_script, 'set_image_data')
    @mock.patch.object(store_utils, 'delete_image_location_from_backend')
    def test_import_image_failed_with_expired_token(
            self, mock_delete_data, mock_set_img_data, mock_create_image):
        image_id = mock.ANY
        locations = ['location']
        image = mock.Mock(image_id=image_id, locations=locations)
        image_repo = mock.Mock()
        image_repo.get.side_effect = [image, exception.NotAuthenticated]
        image_factory = mock.ANY
        task_input = mock.Mock(image_properties=mock.ANY)
        uri = mock.ANY

        mock_create_image.return_value = image
        self.assertRaises(exception.NotAuthenticated,
                          image_import_script.import_image,
                          image_repo, image_factory,
                          task_input, None, uri)
        self.assertEqual(1, mock_set_img_data.call_count)
        mock_delete_data.assert_called_once_with(
            mock_create_image().context, image_id, 'location')

    @mock.patch('oslo_utils.timeutils.StopWatch')
    @mock.patch('glance.common.scripts.utils.get_image_data_iter')
    def test_set_image_data_with_callback(self, mock_gidi, mock_sw):
        data = [b'0' * 60, b'0' * 50, b'0' * 10, b'0' * 150]
        result_data = []
        mock_gidi.return_value = iter(data)
        mock_sw.return_value.expired.side_effect = [False, True, False,
                                                    False]
        image = mock.MagicMock()
        callback = mock.MagicMock()

        def fake_set_data(data_iter, **kwargs):
            for chunk in data_iter:
                result_data.append(chunk)

        image.set_data.side_effect = fake_set_data
        image_import_script.set_image_data(image, 'http://fake', None,
                                           callback=callback)

        mock_gidi.assert_called_once_with('http://fake')
        self.assertEqual(data, result_data)
        # Since we only fired the timer once, only two calls expected
        # for the four reads we did, including the final obligatory one
        callback.assert_has_calls([mock.call(110, 110),
                                   mock.call(160, 270)])



# Copyright (c) 2018-2019 RedHat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import glance_store
from oslo_config import cfg
import webob.exc

import glance.api.v2.discovery
from glance.tests.unit import base
import glance.tests.unit.utils as unit_test_utils


CONF = cfg.CONF


class TestInfoControllers(base.MultiStoreClearingUnitTest):
    def setUp(self):
        super(TestInfoControllers, self).setUp()
        self.controller = glance.api.v2.discovery.InfoController()

    def tearDown(self):
        super(TestInfoControllers, self).tearDown()

    def test_get_stores_with_enabled_backends_empty(self):
        self.config(enabled_backends={})
        req = unit_test_utils.get_fake_request()
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.get_stores,
                          req)

    def test_get_stores(self):
        available_stores = ['cheap', 'fast', 'readonly_store', 'fast-cinder',
                            'fast-rbd', 'reliable']
        req = unit_test_utils.get_fake_request()
        output = self.controller.get_stores(req)
        self.assertIn('stores', output)
        for stores in output['stores']:
            self.assertIn('id', stores)
            self.assertNotIn('weight', stores)
            self.assertIn(stores['id'], available_stores)

    def test_get_stores_read_only_store(self):
        available_stores = ['cheap', 'fast', 'readonly_store', 'fast-cinder',
                            'fast-rbd', 'reliable']
        req = unit_test_utils.get_fake_request()
        output = self.controller.get_stores(req)
        self.assertIn('stores', output)
        for stores in output['stores']:
            self.assertIn('id', stores)
            self.assertIn(stores['id'], available_stores)
            if stores['id'] == 'readonly_store':
                self.assertTrue(stores['read-only'])
            else:
                self.assertIsNone(stores.get('read-only'))

    def test_get_stores_reserved_stores_excluded(self):
        enabled_backends = {
            'fast': 'file',
            'cheap': 'file'
        }
        self.config(enabled_backends=enabled_backends)
        req = unit_test_utils.get_fake_request()
        output = self.controller.get_stores(req)
        self.assertIn('stores', output)
        self.assertEqual(2, len(output['stores']))
        for stores in output["stores"]:
            self.assertFalse(stores["id"].startswith("os_glance_"))

    def test_get_stores_detail(self):
        available_stores = ['cheap', 'fast', 'readonly_store', 'fast-cinder',
                            'fast-rbd', 'reliable']
        available_store_type = ['file', 'file', 'http', 'cinder', 'rbd',
                                'swift']
        req = unit_test_utils.get_fake_request(roles=['admin'])
        output = self.controller.get_stores_detail(req)
        self.assertEqual(len(CONF.enabled_backends), len(output['stores']))
        self.assertIn('stores', output)
        for stores in output['stores']:
            self.assertIn('id', stores)
            self.assertIn(stores['id'], available_stores)
            self.assertIn(stores['type'], available_store_type)
            self.assertIsNotNone(stores['properties'])

    def test_get_stores_detail_properties(self):
        store_attributes = {'rbd': ['chunk_size', 'pool', 'thin_provisioning'],
                            'file': ['data_dir', 'chunk_size',
                                     'thin_provisioning'],
                            'cinder': ['volume_type', 'use_multipath'],
                            'swift': ['container',
                                      'large_object_size',
                                      'large_object_chunk_size'],
                            'http': []}
        req = unit_test_utils.get_fake_request(roles=['admin'])
        output = self.controller.get_stores_detail(req)
        self.assertEqual(len(CONF.enabled_backends), len(output['stores']))
        self.assertIn('stores', output)
        for store in output['stores']:
            actual_attribute = list(store['properties'].keys())
            expected_attribute = store_attributes[store['type']]
            self.assertEqual(actual_attribute, expected_attribute)

    def test_get_stores_detail_with_store_weight(self):
        self.config(weight=100, group='fast')
        self.config(weight=200, group='cheap')
        self.config(weight=300, group='fast-rbd')
        self.config(weight=400, group='fast-cinder')
        self.config(weight=500, group='reliable')

        req = unit_test_utils.get_fake_request(roles=['admin'])
        output = self.controller.get_stores_detail(req)
        self.assertEqual(len(CONF.enabled_backends), len(output['stores']))
        self.assertIn('stores', output)
        for store in output['stores']:
            self.assertIn('weight', store)

    def test_get_stores_detail_non_admin(self):
        req = unit_test_utils.get_fake_request()
        self.assertRaises(webob.exc.HTTPForbidden,
                          self.controller.get_stores_detail,
                          req)

    def test_swift_multitenant_and_conf_file_enabled(self):
        self.config(enabled_backends={'fast-rbd': 'rbd', 'test': 'swift'})
        glance_store.register_store_opts(CONF)
        self.config(default_backend='fast-rbd',
                    group='glance_store')
        self.config(rbd_store_chunk_size=8688388, rbd_store_pool='images',
                    rbd_thin_provisioning=False, group='fast-rbd')
        self.config(swift_store_container='glance',
                    swift_store_large_object_size=524288000,
                    swift_store_large_object_chunk_size=204800000,
                    swift_store_config_file='fake-file.conf',
                    swift_store_multi_tenant=True,
                    group='test')
        glance_store.create_multi_stores(CONF)
        req = unit_test_utils.get_fake_request(roles=['admin'])
        output = self.controller.get_stores_detail(req)
        self.assertNotEqual(len(CONF.enabled_backends), len(output['stores']))
        self.assertNotIn('test',
                         [store.get('id') for store in output['stores']])



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2014 SoftLayer Technologies, Inc.
# Copyright 2015 Mirantis, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
System-level utilities and helper functions.
"""

import errno
import ipaddress

try:
    from eventlet import sleep
except ImportError:
    from time import sleep
from eventlet.green import socket

import functools
import os
import re
import urllib

import glance_store
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import netutils
from oslo_utils import strutils
from webob import exc

from glance.common import exception
from glance.common import timeutils
from glance.common import wsgi
from glance.i18n import _, _LE, _LW

CONF = cfg.CONF

LOG = logging.getLogger(__name__)

# Whitelist of v1 API headers of form x-image-meta-xxx
IMAGE_META_HEADERS = ['x-image-meta-location', 'x-image-meta-size',
                      'x-image-meta-is_public', 'x-image-meta-disk_format',
                      'x-image-meta-container_format', 'x-image-meta-name',
                      'x-image-meta-status', 'x-image-meta-copy_from',
                      'x-image-meta-uri', 'x-image-meta-checksum',
                      'x-image-meta-created_at', 'x-image-meta-updated_at',
                      'x-image-meta-deleted_at', 'x-image-meta-min_ram',
                      'x-image-meta-min_disk', 'x-image-meta-owner',
                      'x-image-meta-store', 'x-image-meta-id',
                      'x-image-meta-protected', 'x-image-meta-deleted',
                      'x-image-meta-virtual_size']

GLANCE_TEST_SOCKET_FD_STR = 'GLANCE_TEST_SOCKET_FD'


def chunkreadable(iter, chunk_size=65536):
    """
    Wrap a readable iterator with a reader yielding chunks of
    a preferred size, otherwise leave iterator unchanged.

    :param iter: an iter which may also be readable
    :param chunk_size: maximum size of chunk
    """
    return chunkiter(iter, chunk_size) if hasattr(iter, 'read') else iter


def chunkiter(fp, chunk_size=65536):
    """
    Return an iterator to a file-like obj which yields fixed size chunks

    :param fp: a file-like object
    :param chunk_size: maximum size of chunk
    """
    while True:
        chunk = fp.read(chunk_size)
        if chunk:
            yield chunk
        else:
            break


def cooperative_iter(iter):
    """
    Return an iterator which schedules after each
    iteration. This can prevent eventlet thread starvation.

    :param iter: an iterator to wrap
    """
    try:
        for chunk in iter:
            sleep(0)
            yield chunk
    except Exception as err:
        with excutils.save_and_reraise_exception():
            msg = _LE("Error: cooperative_iter exception %s") % err
            LOG.error(msg)


def cooperative_read(fd):
    """
    Wrap a file descriptor's read with a partial function which schedules
    after each read. This can prevent eventlet thread starvation.

    :param fd: a file descriptor to wrap
    """
    def readfn(*args):
        result = fd.read(*args)
        sleep(0)
        return result
    return readfn


MAX_COOP_READER_BUFFER_SIZE = 134217728  # 128M seems like a sane buffer limit

CONF.import_group('import_filtering_opts',
                  'glance.async_.flows._internal_plugins')


def normalize_hostname(host):
    """Normalize IP address to standard format or return hostname.

    Uses ipaddress module to validate and normalize IP addresses, rejecting
    encoded formats. For hostnames, requires DNS resolution to ensure they
    are valid and not encoded IP attempts.

    :param host: hostname or IP address
    :returns: normalized IP address, hostname unchanged, or None
    """
    if not host:
        return host

    # NOTE(abhishekk): Try to parse as IPv4. ipaddress module only accepts
    # standard format like 127.0.0.1. It rejects encoded formats like
    # decimal (2130706433), hex (0x7f000001), or octal (017700000001).
    try:
        return str(ipaddress.IPv4Address(host))
    except ValueError:
        pass

    # NOTE(abhishekk): Try to parse as IPv6. ipaddress module only accepts
    # standard IPv6 format and rejects encoded formats.
    try:
        return str(ipaddress.IPv6Address(host))
    except ValueError:
        pass

    # NOTE(abhishekk): Not valid IP address, check as hostname. Reject pure
    # numeric strings like "2130706433" (decimal encoded IP). ipaddress module
    # rejected it, but OS might still resolve using inet_aton() if not blocked.
    if host.isdigit():
        return None

    # NOTE(abhishekk): Reject all numeric strings with dots like "127.1" or
    # "10.1". These are shorthand IP addresses. ipaddress module rejects them
    # because they need 4 octets, but OS may still resolve them. We block to
    # prevent SSRF bypass attacks.
    if all(c.isdigit() or c == '.' for c in host):
        return None

    # NOTE(abhishekk): Add trailing dot to force DNS lookup instead of numeric
    # parsing. This blocks encoded IP formats like 0x7f000001 or 127.0x0.0.1
    # because they fail DNS lookup. Only real hostnames that resolve via DNS
    # are allowed.
    testhost = host
    if not testhost.endswith('.'):
        testhost += '.'

    try:
        socket.getaddrinfo(testhost, 80)
    except socket.gaierror:
        # NOTE(abhishekk): DNS resolution failed, reject the hostname
        return None

    # NOTE(abhishekk): Valid and resolvable hostname, return unchanged
    return host


def validate_import_uri(uri):
    """Validate requested uri for Image Import web-download.

    :param uri: target uri to be validated
    """
    if not uri:
        return False

    parsed_uri = urllib.parse.urlparse(uri)
    scheme = parsed_uri.scheme
    host = parsed_uri.hostname
    port = parsed_uri.port
    wl_schemes = CONF.import_filtering_opts.allowed_schemes
    bl_schemes = CONF.import_filtering_opts.disallowed_schemes
    wl_hosts = CONF.import_filtering_opts.allowed_hosts
    bl_hosts = CONF.import_filtering_opts.disallowed_hosts
    wl_ports = CONF.import_filtering_opts.allowed_ports
    bl_ports = CONF.import_filtering_opts.disallowed_ports

    # NOTE(jokke): Checking if both allowed and disallowed are defined and
    # logging it to inform only allowed will be obeyed.
    if wl_schemes and bl_schemes:
        bl_schemes = []
        LOG.debug("Both allowed and disallowed schemes has been configured. "
                  "Will only process allowed list.")
    if wl_hosts and bl_hosts:
        bl_hosts = []
        LOG.debug("Both allowed and disallowed hosts has been configured. "
                  "Will only process allowed list.")
    if wl_ports and bl_ports:
        bl_ports = []
        LOG.debug("Both allowed and disallowed ports has been configured. "
                  "Will only process allowed list.")

    if not scheme or ((wl_schemes and scheme not in wl_schemes) or
                      parsed_uri.scheme in bl_schemes):
        return False

    normalized_host = normalize_hostname(host)

    if not normalized_host or (
            (wl_hosts and normalized_host not in wl_hosts) or
            normalized_host in bl_hosts):
        return False

    if port and ((wl_ports and port not in wl_ports) or
                 port in bl_ports):
        return False

    return True


class CooperativeReader(object):
    """
    An eventlet thread friendly class for reading in image data.

    When accessing data either through the iterator or the read method
    we perform a sleep to allow a co-operative yield. When there is more than
    one image being uploaded/downloaded this prevents eventlet thread
    starvation, ie allows all threads to be scheduled periodically rather than
    having the same thread be continuously active.
    """
    def __init__(self, fd):
        """
        :param fd: Underlying image file object
        """
        self.fd = fd
        self.iterator = None
        # NOTE(markwash): if the underlying supports read(), overwrite the
        # default iterator-based implementation with cooperative_read which
        # is more straightforward
        if hasattr(fd, 'read'):
            self.read = cooperative_read(fd)
        else:
            self.iterator = None
            self.buffer = b''
            self.position = 0

    def read(self, length=None):
        """Return the requested amount of bytes, fetching the next chunk of
        the underlying iterator when needed.

        This is replaced with cooperative_read in __init__ if the underlying
        fd already supports read().
        """
        if length is None:
            if len(self.buffer) - self.position > 0:
                # if no length specified but some data exists in buffer,
                # return that data and clear the buffer
                result = self.buffer[self.position:]
                self.buffer = b''
                self.position = 0
                return bytes(result)
            else:
                # otherwise read the next chunk from the underlying iterator
                # and return it as a whole. Reset the buffer, as subsequent
                # calls may specify the length
                try:
                    if self.iterator is None:
                        self.iterator = self.__iter__()
                    return next(self.iterator)
                except StopIteration:
                    return b''
                finally:
                    self.buffer = b''
                    self.position = 0
        else:
            result = bytearray()
            while len(result) < length:
                if self.position < len(self.buffer):
                    to_read = length - len(result)
                    chunk = self.buffer[self.position:self.position + to_read]
                    result.extend(chunk)

                    # This check is here to prevent potential OOM issues if
                    # this code is called with unreasonably high values of read
                    # size. Currently it is only called from the HTTP clients
                    # of Glance backend stores, which use httplib for data
                    # streaming, which has readsize hardcoded to 8K, so this
                    # check should never fire. Regardless it still worths to
                    # make the check, as the code may be reused somewhere else.
                    if len(result) >= MAX_COOP_READER_BUFFER_SIZE:
                        raise exception.LimitExceeded()
                    self.position += len(chunk)
                else:
                    try:
                        if self.iterator is None:
                            self.iterator = self.__iter__()
                        self.buffer = next(self.iterator)
                        self.position = 0
                    except StopIteration:
                        self.buffer = b''
                        self.position = 0
                        return bytes(result)
            return bytes(result)

    def __iter__(self):
        return cooperative_iter(self.fd.__iter__())


class LimitingReader(object):
    """
    Reader designed to fail when reading image data past the configured
    allowable amount.
    """
    def __init__(self, data, limit,
                 exception_class=exception.ImageSizeLimitExceeded):
        """
        :param data: Underlying image data object
        :param limit: maximum number of bytes the reader should allow
        :param exception_class: Type of exception to be raised
        """
        self.data = data
        self.limit = limit
        self.bytes_read = 0
        self.exception_class = exception_class

    def __iter__(self):
        for chunk in self.data:
            self.bytes_read += len(chunk)
            if self.bytes_read > self.limit:
                raise self.exception_class()
            else:
                yield chunk

    def read(self, i):
        result = self.data.read(i)
        self.bytes_read += len(result)
        if self.bytes_read > self.limit:
            raise self.exception_class()
        return result


def image_meta_to_http_headers(image_meta):
    """
    Returns a set of image metadata into a dict
    of HTTP headers that can be fed to either a Webob
    Request object or an httplib.HTTP(S)Connection object

    :param image_meta: Mapping of image metadata
    """
    headers = {}
    for k, v in image_meta.items():
        if v is not None:
            if k == 'properties':
                for pk, pv in v.items():
                    if pv is not None:
                        headers["x-image-meta-property-%s"
                                % pk.lower()] = str(pv)
            else:
                headers["x-image-meta-%s" % k.lower()] = str(v)
    return headers


def get_image_meta_from_headers(response):
    """
    Processes HTTP headers from a supplied response that
    match the x-image-meta and x-image-meta-property and
    returns a mapping of image metadata and properties

    :param response: Response to process
    """
    result = {}
    properties = {}

    if hasattr(response, 'getheaders'):  # httplib.HTTPResponse
        headers = response.getheaders()
    else:  # webob.Response
        headers = response.headers.items()

    for key, value in headers:
        key = str(key.lower())
        if key.startswith('x-image-meta-property-'):
            field_name = key[len('x-image-meta-property-'):].replace('-', '_')
            properties[field_name] = value or None
        elif key.startswith('x-image-meta-'):
            field_name = key[len('x-image-meta-'):].replace('-', '_')
            if 'x-image-meta-' + field_name not in IMAGE_META_HEADERS:
                msg = _("Bad header: %(header_name)s") % {'header_name': key}
                raise exc.HTTPBadRequest(msg, content_type="text/plain")
            result[field_name] = value or None
    result['properties'] = properties

    for key, nullable in [('size', False), ('min_disk', False),
                          ('min_ram', False), ('virtual_size', True)]:
        if key in result:
            try:
                result[key] = int(result[key])
            except ValueError:
                if nullable and result[key] == str(None):
                    result[key] = None
                else:
                    extra = (_("Cannot convert image %(key)s '%(value)s' "
                               "to an integer.")
                             % {'key': key, 'value': result[key]})
                    raise exception.InvalidParameterValue(value=result[key],
                                                          param=key,
                                                          extra_msg=extra)
            if result[key] is not None and result[key] < 0:
                extra = _('Cannot be a negative value.')
                raise exception.InvalidParameterValue(value=result[key],
                                                      param=key,
                                                      extra_msg=extra)

    for key in ('is_public', 'deleted', 'protected'):
        if key in result:
            result[key] = strutils.bool_from_string(result[key])
    return result


def create_mashup_dict(image_meta):
    """
    Returns a dictionary-like mashup of the image core properties
    and the image custom properties from given image metadata.

    :param image_meta: metadata of image with core and custom properties
    """

    d = {}
    for key, value in image_meta.items():
        if isinstance(value, dict):
            for subkey, subvalue in create_mashup_dict(value).items():
                if subkey not in image_meta:
                    d[subkey] = subvalue
        else:
            d[key] = value

    return d


def safe_mkdirs(path):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def mutating(func):
    """Decorator to enforce read-only logic"""
    @functools.wraps(func)
    def wrapped(self, req, *args, **kwargs):
        if req.context.read_only:
            msg = "Read-only access"
            LOG.debug(msg)
            raise exc.HTTPForbidden(msg, request=req,
                                    content_type="text/plain")
        return func(self, req, *args, **kwargs)
    return wrapped


def setup_remote_pydev_debug(host, port):
    error_msg = _LE('Error setting up the debug environment. Verify that the'
                    ' option pydev_worker_debug_host is pointing to a valid '
                    'hostname or IP on which a pydev server is listening on'
                    ' the port indicated by pydev_worker_debug_port.')

    try:
        try:
            from pydev import pydevd
        except ImportError:
            import pydevd

        pydevd.settrace(host,
                        port=port,
                        stdoutToServer=True,
                        stderrToServer=True)
        return True
    except Exception:
        with excutils.save_and_reraise_exception():
            LOG.exception(error_msg)


def get_test_suite_socket():
    global GLANCE_TEST_SOCKET_FD_STR
    if GLANCE_TEST_SOCKET_FD_STR in os.environ:
        fd = int(os.environ[GLANCE_TEST_SOCKET_FD_STR])
        sock = socket.fromfd(fd, socket.AF_INET, socket.SOCK_STREAM)
        sock.listen(CONF.backlog)
        del os.environ[GLANCE_TEST_SOCKET_FD_STR]
        os.close(fd)
        return sock
    return None


def is_valid_hostname(hostname):
    """Verify whether a hostname (not an FQDN) is valid."""
    return re.match('^[a-zA-Z0-9-]+$', hostname) is not None


def is_valid_fqdn(fqdn):
    """Verify whether a host is a valid FQDN."""
    return re.match(r'^[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$', fqdn) is not None


def parse_valid_host_port(host_port):
    """
    Given a "host:port" string, attempts to parse it as intelligently as
    possible to determine if it is valid. This includes IPv6 [host]:port form,
    IPv4 ip:port form, and hostname:port or fqdn:port form.

    Invalid inputs will raise a ValueError, while valid inputs will return
    a (host, port) tuple where the port will always be of type int.
    """

    try:
        try:
            host, port = netutils.parse_host_port(host_port)
        except Exception:
            raise ValueError(_('Host and port "%s" is not valid.') % host_port)

        if not netutils.is_valid_port(port):
            raise ValueError(_('Port "%s" is not valid.') % port)

        # First check for valid IPv6 and IPv4 addresses, then a generic
        # hostname. Failing those, if the host includes a period, then this
        # should pass a very generic FQDN check. The FQDN check for letters at
        # the tail end will weed out any hilariously absurd IPv4 addresses.

        if not (netutils.is_valid_ipv6(host) or netutils.is_valid_ipv4(host) or
                is_valid_hostname(host) or is_valid_fqdn(host)):
            raise ValueError(_('Host "%s" is not valid.') % host)

    except Exception as ex:
        raise ValueError(_('%s '
                           'Please specify a host:port pair, where host is an '
                           'IPv4 address, IPv6 address, hostname, or FQDN. If '
                           'using an IPv6 address, enclose it in brackets '
                           'separately from the port (i.e., '
                           '"[fe80::a:b:c]:9876").') % ex)

    return (host, int(port))


try:
    REGEX_4BYTE_UNICODE = re.compile('[\U00010000-\U0010ffff]')
except re.error:
    # UCS-2 build case
    REGEX_4BYTE_UNICODE = re.compile('[\uD800-\uDBFF][\uDC00-\uDFFF]')


def no_4byte_params(f):
    """
    Checks that no 4 byte unicode characters are allowed
    in dicts' keys/values and string's parameters
    """
    def wrapper(*args, **kwargs):

        def _is_match(some_str):
            return (
                isinstance(some_str, str) and
                REGEX_4BYTE_UNICODE.findall(some_str) != []
            )

        def _check_dict(data_dict):
            # a dict of dicts has to be checked recursively
            for key, value in data_dict.items():
                if isinstance(value, dict):
                    _check_dict(value)
                else:
                    if _is_match(key):
                        msg = _("Property names can't contain 4 byte unicode.")
                        raise exception.Invalid(msg)
                    if _is_match(value):
                        msg = (_("%s can't contain 4 byte unicode characters.")
                               % key.title())
                        raise exception.Invalid(msg)

        for data_dict in [arg for arg in args if isinstance(arg, dict)]:
            _check_dict(data_dict)
        # now check args for str values
        for arg in args:
            if _is_match(arg):
                msg = _("Param values can't contain 4 byte unicode.")
                raise exception.Invalid(msg)
        # check kwargs as well, as params are passed as kwargs via
        # registry calls
        _check_dict(kwargs)
        return f(*args, **kwargs)
    return wrapper


def stash_conf_values():
    """
    Make a copy of some of the current global CONF's settings.
    Allows determining if any of these values have changed
    when the config is reloaded.
    """
    conf = {
        'bind_host': CONF.bind_host,
        'bind_port': CONF.bind_port,
        'backlog': CONF.backlog,
    }

    return conf


def split_filter_op(expression):
    """Split operator from threshold in an expression.
    Designed for use on a comparative-filtering query field.
    When no operator is found, default to an equality comparison.

    :param expression: the expression to parse

    :returns: a tuple (operator, threshold) parsed from expression
    """
    left, sep, right = expression.partition(':')
    if sep:
        # If the expression is a date of the format ISO 8601 like
        # CCYY-MM-DDThh:mm:ss+hh:mm and has no operator, it should
        # not be partitioned, and a default operator of eq should be
        # assumed.
        try:
            timeutils.parse_isotime(expression)
            op = 'eq'
            threshold = expression
        except ValueError:
            op = left
            threshold = right
    else:
        op = 'eq'  # default operator
        threshold = left

    # NOTE stevelle decoding escaped values may be needed later
    return op, threshold


def validate_quotes(value):
    """Validate filter values

    Validation opening/closing quotes in the expression.
    """
    open_quotes = True
    for i in range(len(value)):
        if value[i] == '"':
            if i and value[i - 1] == '\\':
                continue
            if open_quotes:
                if i and value[i - 1] != ',':
                    msg = _("Invalid filter value %s. There is no comma "
                            "before opening quotation mark.") % value
                    raise exception.InvalidParameterValue(message=msg)
            else:
                if i + 1 != len(value) and value[i + 1] != ",":
                    msg = _("Invalid filter value %s. There is no comma "
                            "after closing quotation mark.") % value
                    raise exception.InvalidParameterValue(message=msg)
            open_quotes = not open_quotes
    if not open_quotes:
        msg = _("Invalid filter value %s. The quote is not closed.") % value
        raise exception.InvalidParameterValue(message=msg)


def split_filter_value_for_quotes(value):
    """Split filter values

    Split values by commas and quotes for 'in' operator, according api-wg.
    """
    validate_quotes(value)
    tmp = re.compile(r'''
        "(                 # if found a double-quote
           [^\"\\]*        # take characters either non-quotes or backslashes
           (?:\\.          # take backslashes and character after it
            [^\"\\]*)*     # take characters either non-quotes or backslashes
         )                 # before double-quote
        ",?                # a double-quote with comma maybe
        | ([^,]+),?        # if not found double-quote take any non-comma
                           # characters with comma maybe
        | ,                # if we have only comma take empty string
        ''', re.VERBOSE)
    return [val[0] or val[1] for val in re.findall(tmp, value)]


def evaluate_filter_op(value, operator, threshold):
    """Evaluate a comparison operator.
    Designed for use on a comparative-filtering query field.

    :param value: evaluated against the operator, as left side of expression
    :param operator: any supported filter operation
    :param threshold: to compare value against, as right side of expression

    :raises InvalidFilterOperatorValue: if an unknown operator is provided

    :returns: boolean result of applied comparison

    """
    if operator == 'gt':
        return value > threshold
    elif operator == 'gte':
        return value >= threshold
    elif operator == 'lt':
        return value < threshold
    elif operator == 'lte':
        return value <= threshold
    elif operator == 'neq':
        return value != threshold
    elif operator == 'eq':
        return value == threshold

    msg = _("Unable to filter on a unknown operator.")
    raise exception.InvalidFilterOperatorValue(msg)


def _get_available_stores():
    available_stores = CONF.enabled_backends
    stores = []
    # Remove reserved stores from the available stores list
    for store in available_stores:
        # NOTE (abhishekk): http store is readonly and should be
        # excluded from the list.
        if available_stores[store] == 'http':
            continue
        if store not in wsgi.RESERVED_STORES:
            stores.append(store)

    return stores


def get_stores_from_request(req, body):
    """Processes a supplied request and extract stores from it

    :param req: request to process
    :param body: request body

    :raises glance_store.UnknownScheme:  if a store is not valid
    :return: a list of stores
    """
    if body.get('all_stores', False):
        if 'stores' in body or 'x-image-meta-store' in req.headers:
            msg = _("All_stores parameter can't be used with "
                    "x-image-meta-store header or stores parameter")
            raise exc.HTTPBadRequest(explanation=msg)
        stores = _get_available_stores()
    else:
        try:
            stores = body['stores']
        except KeyError:
            stores = [req.headers.get('x-image-meta-store',
                                      CONF.glance_store.default_backend)]
        else:
            if 'x-image-meta-store' in req.headers:
                msg = _("Stores parameter and x-image-meta-store header can't "
                        "be both specified")
                raise exc.HTTPBadRequest(explanation=msg)
    # Validate each store
    for store in stores:
        glance_store.get_store_from_store_identifier(store)
    return stores


def sort_image_locations(locations):
    if not CONF.enabled_backends:
        return locations

    def get_store_weight(location):
        store_id = location['metadata'].get('store')
        if not store_id:
            return 0
        try:
            store = glance_store.get_store_from_store_identifier(store_id)
        except glance_store.exceptions.UnknownScheme:
            msg = (_LW("Unable to find store '%s', returning "
                       "default weight '0'") % store_id)
            LOG.warning(msg)
            return 0
        return store.weight if store is not None else 0

    sorted_locations = sorted(locations, key=get_store_weight, reverse=True)
    scrubbed = []
    for loc in sorted_locations:
        scrubbed.append({'store_name': loc["metadata"].get("store")})

    LOG.debug(('Sorted locations: %s'), scrubbed)
    return sorted_locations


def is_http_store_configured(url):
    if not url.startswith("http"):
        return False

    enabled_backends = CONF.enabled_backends
    if enabled_backends:
        return 'http' in enabled_backends.values()
    else:
        return 'http' in CONF.glance_store.stores



# Copyright 2015 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import http.client as http

import glance_store
from oslo_log import log as logging
import webob.exc

from glance.api import policy
from glance.api.v2 import policy as api_policy
from glance.common import exception
from glance.common import utils
from glance.common import wsgi
import glance.db
import glance.gateway
from glance.i18n import _LI
import glance.notifier


LOG = logging.getLogger(__name__)


class ImageActionsController(object):
    def __init__(self, db_api=None, policy_enforcer=None, notifier=None,
                 store_api=None):
        self.db_api = db_api or glance.db.get_api()
        self.policy = policy_enforcer or policy.Enforcer()
        self.notifier = notifier or glance.notifier.Notifier()
        self.store_api = store_api or glance_store
        self.gateway = glance.gateway.Gateway(self.db_api, self.store_api,
                                              self.notifier, self.policy)

    @utils.mutating
    def deactivate(self, req, image_id):
        image_repo = self.gateway.get_repo(req.context)
        try:
            # FIXME(danms): This will still enforce the get_image policy
            # which we don't want
            image = image_repo.get(image_id)

            # NOTE(abhishekk): This is the right place to check whether user
            # have permission to deactivate the image and remove the policy
            # check later from the policy layer.
            api_pol = api_policy.ImageAPIPolicy(req.context, image,
                                                self.policy)
            api_pol.deactivate_image()

            status = image.status
            image.deactivate()
            # not necessary to change the status if it's already 'deactivated'
            if status == 'active':
                image_repo.save(image, from_state='active')
            LOG.info(_LI("Image %s is deactivated"), image_id)
        except exception.NotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.msg)
        except exception.Forbidden as e:
            LOG.debug("User not permitted to deactivate image '%s'", image_id)
            raise webob.exc.HTTPForbidden(explanation=e.msg)
        except exception.InvalidImageStatusTransition as e:
            raise webob.exc.HTTPBadRequest(explanation=e.msg)

    @utils.mutating
    def reactivate(self, req, image_id):
        image_repo = self.gateway.get_repo(req.context)
        try:
            # FIXME(danms): This will still enforce the get_image policy
            # which we don't want
            image = image_repo.get(image_id)

            # NOTE(abhishekk): This is the right place to check whether user
            # have permission to reactivate the image and remove the policy
            # check later from the policy layer.
            api_pol = api_policy.ImageAPIPolicy(req.context, image,
                                                self.policy)
            api_pol.reactivate_image()

            status = image.status
            image.reactivate()
            # not necessary to change the status if it's already 'active'
            if status == 'deactivated':
                image_repo.save(image, from_state='deactivated')
            LOG.info(_LI("Image %s is reactivated"), image_id)
        except exception.NotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.msg)
        except exception.Forbidden as e:
            LOG.debug("User not permitted to reactivate image '%s'", image_id)
            raise webob.exc.HTTPForbidden(explanation=e.msg)
        except exception.InvalidImageStatusTransition as e:
            raise webob.exc.HTTPBadRequest(explanation=e.msg)


class ResponseSerializer(wsgi.JSONResponseSerializer):

    def deactivate(self, response, result):
        response.status_int = http.NO_CONTENT

    def reactivate(self, response, result):
        response.status_int = http.NO_CONTENT


def create_resource():
    """Image data resource factory method"""
    deserializer = None
    serializer = ResponseSerializer()
    controller = ImageActionsController()
    return wsgi.Resource(controller, deserializer, serializer)



# Copyright 2021 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import oslo_policy.policy
from oslo_utils.fixture import uuidsentinel as uuids

from glance.api import policy
from glance.tests import functional


class TestImagesPolicy(functional.SynchronousAPIBase):
    def setUp(self):
        super(TestImagesPolicy, self).setUp()
        self.policy = policy.Enforcer(suppress_deprecation_warnings=True)

    def set_policy_rules(self, rules):
        self.policy.set_rules(
            oslo_policy.policy.Rules.from_dict(rules),
            overwrite=True)

    def start_server(self):
        with mock.patch.object(policy, 'Enforcer') as mock_enf:
            mock_enf.return_value = self.policy
            super(TestImagesPolicy, self).start_server()

    def test_image_update_basic(self):
        self.start_server()
        image_id = self._create_and_upload()

        # First make sure image update works with the default policy
        resp = self.api_patch('/v2/images/%s' % image_id,
                              {'op': 'add',
                               'path': '/mykey1',
                               'value': 'foo'})
        self.assertEqual(200, resp.status_code, resp.text)
        self.assertEqual(
            'foo',
            self.api_get('/v2/images/%s' % image_id).json['mykey1'])

        # Now disable modify_image permissions and make sure any other
        # attempts fail
        self.set_policy_rules({'get_image': '',
                               'modify_image': '!'})

        # Add should fail
        resp = self.api_patch('/v2/images/%s' % image_id,
                              {'op': 'add',
                               'path': '/mykey2',
                               'value': 'foo'})
        self.assertEqual(403, resp.status_code)
        self.assertNotIn(
            'mykey2',
            self.api_get('/v2/images/%s' % image_id).json)

        # Replace should fail, old value should persist
        resp = self.api_patch('/v2/images/%s' % image_id,
                              {'op': 'replace',
                               'path': '/mykey1',
                               'value': 'bar'})
        self.assertEqual(403, resp.status_code)
        self.assertEqual(
            'foo',
            self.api_get('/v2/images/%s' % image_id).json['mykey1'])

        # Remove should fail, old value should persist
        resp = self.api_patch('/v2/images/%s' % image_id,
                              {'op': 'remove',
                               'path': '/mykey1'})
        self.assertEqual(403, resp.status_code)
        self.assertEqual(
            'foo',
            self.api_get('/v2/images/%s' % image_id).json['mykey1'])

        # Now disable get_image permissions and we should get a 404
        # instead of a 403 when trying to do the same operation as above.
        # Remove should fail, old value should persist
        self.set_policy_rules({'get_image': '!',
                               'modify_image': '!'})
        resp = self.api_patch('/v2/images/%s' % image_id,
                              {'op': 'remove',
                               'path': '/mykey1'})
        self.assertEqual(404, resp.status_code)

    @mock.patch('glance.location._check_image_location', new=lambda *a: 0)
    @mock.patch('glance.location.ImageRepoProxy._set_acls', new=lambda *a: 0)
    def test_image_update_locations(self):
        self.config(show_multiple_locations=True)
        self.start_server()
        image_id = self._create_and_upload()

        # First make sure we can add and delete locations
        resp = self.api_patch('/v2/images/%s' % image_id,
                              {'op': 'add',
                               'path': '/locations/0',
                               'value': {'url': 'http://foo.bar',
                                         'metadata': {}}})
        self.assertEqual(200, resp.status_code, resp.text)
        self.assertEqual(2,
                         len(self.api_get(
                             '/v2/images/%s' % image_id).json['locations']))
        self.assertEqual(
            'http://foo.bar',
            self.api_get(
                '/v2/images/%s' % image_id).json['locations'][1]['url'])

        resp = self.api_patch('/v2/images/%s' % image_id,
                              {'op': 'remove',
                               'path': '/locations/0'})
        self.assertEqual(200, resp.status_code, resp.text)
        self.assertEqual(1,
                         len(self.api_get(
                             '/v2/images/%s' % image_id).json['locations']))

        # Add another while we still can so we can try to delete it below
        resp = self.api_patch('/v2/images/%s' % image_id,
                              {'op': 'add',
                               'path': '/locations/0',
                               'value': {'url': 'http://foo.baz',
                                         'metadata': {}}})
        self.assertEqual(200, resp.status_code, resp.text)
        self.assertEqual(2,
                         len(self.api_get(
                             '/v2/images/%s' % image_id).json['locations']))

        # Now disable set/delete_image_location permissions and make
        # sure any other attempts fail
        self.set_policy_rules({'get_image': '',
                               'get_image_location': '',
                               'set_image_location': '!',
                               'delete_image_location': '!'})

        # Make sure we cannot delete the above or add another
        resp = self.api_patch('/v2/images/%s' % image_id,
                              {'op': 'remove',
                               'path': '/locations/0'})
        self.assertEqual(403, resp.status_code, resp.text)
        self.assertEqual(2,
                         len(self.api_get(
                             '/v2/images/%s' % image_id).json['locations']))

        resp = self.api_patch('/v2/images/%s' % image_id,
                              {'op': 'add',
                               'path': '/locations/0',
                               'value': {'url': 'http://foo.baz',
                                         'metadata': {}}})
        self.assertEqual(403, resp.status_code, resp.text)
        self.assertEqual(2,
                         len(self.api_get(
                             '/v2/images/%s' % image_id).json['locations']))

    def test_image_get(self):
        self.start_server()

        image_id = self._create_and_upload()

        # Make sure we can get the image
        image = self.api_get('/v2/images/%s' % image_id).json
        self.assertEqual(image_id, image['id'])

        # Make sure we can list the image
        images = self.api_get('/v2/images').json['images']
        self.assertEqual(1, len(images))
        self.assertEqual(image_id, images[0]['id'])

        # Now disable get_images but allow get_image
        self.set_policy_rules({'get_images': '!',
                               'get_image': ''})

        # We should not be able to list, but still fetch the image by id
        resp = self.api_get('/v2/images')
        self.assertEqual(403, resp.status_code)
        image = self.api_get('/v2/images/%s' % image_id).json
        self.assertEqual(image_id, image['id'])

        # Now disable get_image but allow get_images
        self.set_policy_rules({'get_images': '',
                               'get_image': '!'})

        # We should be able to list, but not actually see the image in the list
        images = self.api_get('/v2/images').json['images']
        self.assertEqual(0, len(images))
        resp = self.api_get('/v2/images/%s' % image_id)
        self.assertEqual(404, resp.status_code)

        # Now disable both get_image and get_images
        self.set_policy_rules({'get_images': '!',
                               'get_image': '!'})

        # We should not be able to list or fetch by id
        resp = self.api_get('/v2/images')
        self.assertEqual(403, resp.status_code)
        resp = self.api_get('/v2/images/%s' % image_id)
        self.assertEqual(404, resp.status_code)

    def test_image_create(self):
        self.start_server()

        # Make sure we can create an image
        self.assertEqual(201, self._create().status_code)

        # Now disable add_image and make sure we get 403
        self.set_policy_rules({'add_image': '!'})

        self.assertEqual(403, self._create().status_code)

    def test_image_create_by_another(self):
        self.start_server()

        # NOTE(danms): There is no policy override in this test,
        # specifically to test that the defaults (for rbac and
        # non-rbac) properly catch the attempt by a non-admin to
        # create an image owned by someone else.

        image = {'name': 'foo',
                 'container_format': 'bare',
                 'disk_format': 'raw',
                 'owner': 'someoneelse'}
        resp = self.api_post('/v2/images',
                             json=image,
                             headers={'X-Roles': 'member'})
        # Make sure we get the expected owner-specific error message
        self.assertIn("You are not permitted to create images "
                      "owned by 'someoneelse'", resp.text)

    def test_image_delete(self):
        self.start_server()

        image_id = self._create_and_upload()

        # Make sure we can delete the image
        resp = self.api_delete('/v2/images/%s' % image_id)
        self.assertEqual(204, resp.status_code)

        # Make sure it is really gone
        resp = self.api_get('/v2/images/%s' % image_id)
        self.assertEqual(404, resp.status_code)

        # Make sure we get a 404 trying to delete a non-existent image
        resp = self.api_delete('/v2/images/%s' % image_id)
        self.assertEqual(404, resp.status_code)

        image_id = self._create_and_upload()

        # Now disable delete permissions, but allow get_image
        self.set_policy_rules({'get_image': '',
                               'delete_image': '!'})

        # Make sure delete returns 403 because we can see the image,
        # just not delete it
        resp = self.api_delete('/v2/images/%s' % image_id)
        self.assertEqual(403, resp.status_code)

        # Now disable delete permissions, including get_image
        self.set_policy_rules({'get_image': '!',
                               'delete_image': '!'})

        # Make sure delete returns 404 because we can not see nor
        # delete it
        resp = self.api_delete('/v2/images/%s' % image_id)
        self.assertEqual(404, resp.status_code)

        # Now allow delete, but disallow get_image, just to prove that
        # you do not need get_image in order to be granted delete, and
        # that we only use it for error code determination if
        # permission is denied.
        self.set_policy_rules({'get_image': '!',
                               'delete_image': ''})

        # Make sure delete returns 204 because even though we can not
        # see the image, we can delete it
        resp = self.api_delete('/v2/images/%s' % image_id)
        self.assertEqual(204, resp.status_code)

    def test_image_upload(self):
        self.start_server()

        # Make sure we can upload the image
        self._create_and_upload(expected_code=204)

        # Now disable upload permissions, but allow get_image
        self.set_policy_rules({
            'add_image': '',
            'get_image': '',
            'upload_image': '!'
        })

        # Make sure upload returns 403 because we can see the image,
        # just not upload data to it
        self._create_and_upload(expected_code=403)

        # Now disable upload permissions, including get_image
        self.set_policy_rules({
            'add_image': '',
            'get_image': '!',
            'upload_image': '!',
        })

        # Make sure upload returns 404 because we can not see nor
        # upload data to it
        self._create_and_upload(expected_code=404)

        # Now allow upload, but disallow get_image, just to prove that
        # you do not need get_image in order to be granted upload, and
        # that we only use it for error code determination if
        # permission is denied.
        self.set_policy_rules({
            'add_image': '',
            'get_image': '!',
            'upload_image': ''})

        # Make sure upload returns 204 because even though we can not
        # see the image, we can upload data to it
        self._create_and_upload(expected_code=204)

    def test_image_download(self):
        # NOTE(abhishekk): These tests are running without cache middleware
        self.start_server()
        image_id = self._create_and_upload()

        # First make sure we can download image
        path = '/v2/images/%s/file' % image_id
        response = self.api_get(path)
        self.assertEqual(200, response.status_code)
        self.assertEqual('IMAGEDATA', response.text)

        # Now disable download permissions, but allow get_image
        self.set_policy_rules({
            'get_image': '',
            'download_image': '!'
        })

        # Make sure download returns 403 because we can see the image,
        # just not download it
        response = self.api_get(path)
        self.assertEqual(403, response.status_code)

        # Now disable download permissions, including get_image
        self.set_policy_rules({
            'get_image': '!',
            'download_image': '!',
        })

        # Make sure download returns 404 because we can not see nor
        # download it
        response = self.api_get(path)
        self.assertEqual(404, response.status_code)

        # Now allow download, but disallow get_image, just to prove that
        # you do not need get_image in order to be granted download, and
        # that we only use it for error code determination if
        # permission is denied.
        self.set_policy_rules({
            'get_image': '!',
            'download_image': ''})

        # Make sure download returns 200 because even though we can not
        # see the image, we can download it
        response = self.api_get(path)
        self.assertEqual(200, response.status_code)
        self.assertEqual('IMAGEDATA', response.text)

    def test_image_stage(self):
        self.start_server()
        # First make sure we can perform staging operation
        self._create_and_stage(expected_code=204)

        # Now disable get_image permissions, but allow modify_image
        # should return 204 as well, means even if we can not see
        # image details, we can stage data for it.
        self.set_policy_rules({
            'get_image': '!',
            'modify_image': '',
            'add_image': ''
        })
        self._create_and_stage(expected_code=204)

        # Now allow get_image and disable modify_image should return 403
        self.set_policy_rules({
            'get_image': '',
            'modify_image': '!',
            'add_image': ''
        })
        self._create_and_stage(expected_code=403)

        # Now disabling both permissions will return 404
        self.set_policy_rules({
            'get_image': '!',
            'modify_image': '!',
            'add_image': ''
        })
        self._create_and_stage(expected_code=404)

        # create shared visibility image and stage by 2nd project should
        # return 404 until it is actually shared with that project.
        self.set_policy_rules({
            'get_image': '',
            'modify_image': '!',
            'add_image': '',
            'add_member': ''
        })
        resp = self.api_post('/v2/images',
                             json={'name': 'foo',
                                   'container_format': 'bare',
                                   'disk_format': 'raw',
                                   'visibility': 'shared'})
        self.assertEqual(201, resp.status_code, resp.text)
        image = resp.json
        # Now stage data using another project details
        headers = self._headers({
            'X-Project-Id': 'fake-tenant-id',
            'Content-Type': 'application/octet-stream'
        })
        resp = self.api_put(
            '/v2/images/%s/stage' % image['id'],
            headers=headers,
            data=b'IMAGEDATA')
        self.assertEqual(404, resp.status_code)

        # Now share image with another project and then staging
        # data by that project should return 403
        path = '/v2/images/%s/members' % image['id']
        data = {
            'member': uuids.random_member
        }
        response = self.api_post(path, json=data)
        member = response.json
        self.assertEqual(200, response.status_code)
        self.assertEqual(image['id'], member['image_id'])

        # Now stage data using another project details
        headers = self._headers({
            'X-Project-Id': uuids.random_member,
            'X-Roles': 'member',
            'Content-Type': 'application/octet-stream'
        })
        resp = self.api_put(
            '/v2/images/%s/stage' % image['id'],
            headers=headers,
            data=b'IMAGEDATA')
        self.assertEqual(403, resp.status_code)

    def test_image_deactivate(self):
        self.start_server()

        image_id = self._create_and_upload()

        # Make sure we can deactivate the image
        resp = self.api_post('/v2/images/%s/actions/deactivate' % image_id)
        self.assertEqual(204, resp.status_code)

        # Make sure it is really deactivated
        resp = self.api_get('/v2/images/%s' % image_id)
        self.assertEqual('deactivated', resp.json['status'])

        # Create another image
        image_id = self._create_and_upload()

        # Now disable deactivate permissions, but allow get_image
        self.set_policy_rules({'get_image': '',
                               'deactivate': '!'})

        # Make sure deactivate returns 403 because we can see the image,
        # just not deactivate it
        resp = self.api_post('/v2/images/%s/actions/deactivate' % image_id)
        self.assertEqual(403, resp.status_code)

        # Now disable deactivate permissions, including get_image
        self.set_policy_rules({'get_image': '!',
                               'deactivate': '!'})

        # Make sure deactivate returns 404 because we can not see nor
        # reactivate it
        resp = self.api_post('/v2/images/%s/actions/deactivate' % image_id)
        self.assertEqual(404, resp.status_code)

        # Now allow deactivate, but disallow get_image, just to prove that
        # you do not need get_image in order to be granted deactivate, and
        # that we only use it for error code determination if
        # permission is denied.
        self.set_policy_rules({'get_image': '!',
                               'deactivate': ''})

        # Make sure deactivate returns 204 because even though we can not
        # see the image, we can deactivate it
        resp = self.api_post('/v2/images/%s/actions/deactivate' % image_id)
        self.assertEqual(204, resp.status_code)

        # Make sure you can not deactivate image using non-admin role of
        # different project
        self.set_policy_rules({
            'get_image': '',
            'modify_image': '',
            'add_image': '',
            'upload_image': '',
            'add_member': '',
            'deactivate': '',
            'publicize_image': '',
            'communitize_image': ''
        })
        headers = self._headers({
            'X-Project-Id': 'fake-project-id',
            'X-Roles': 'member'
        })
        for visibility in ('community', 'shared', 'private', 'public'):
            image_id = self._create_and_upload(visibility=visibility)
            resp = self.api_post(
                '/v2/images/%s/actions/deactivate' % image_id, headers=headers)
            # 'shared' image will return 404 until it is not shared with
            # project accessing it
            if visibility == 'shared':
                self.assertEqual(404, resp.status_code)
                # Now lets share the image and try to deactivate it
                share_path = '/v2/images/%s/members' % image_id
                data = {
                    'member': 'fake-project-id'
                }
                response = self.api_post(share_path, json=data)
                member = response.json
                self.assertEqual(200, response.status_code)
                self.assertEqual(image_id, member['image_id'])

                # Now ensure deactivating image by another tenant will
                # return 403
                resp = self.api_post(
                    '/v2/images/%s/actions/deactivate' % image_id,
                    headers=headers)
                self.assertEqual(403, resp.status_code)
            elif visibility == 'private':
                # private image will also return 404 as it is not visible
                self.assertEqual(404, resp.status_code)
            else:
                # public and community visibility will return 403
                self.assertEqual(403, resp.status_code)

    def test_image_reactivate(self):
        self.start_server()

        image_id = self._create_and_upload()

        # deactivate the image
        resp = self.api_post('/v2/images/%s/actions/deactivate' % image_id)
        self.assertEqual(204, resp.status_code)

        # Make sure it is really deactivated
        resp = self.api_get('/v2/images/%s' % image_id)
        self.assertEqual('deactivated', resp.json['status'])

        # Make sure you can reactivate the image
        resp = self.api_post('/v2/images/%s/actions/reactivate' % image_id)
        self.assertEqual(204, resp.status_code)

        # Make sure it is really reactivated
        resp = self.api_get('/v2/images/%s' % image_id)
        self.assertEqual('active', resp.json['status'])

        # Deactivate it again to test further scenarios
        resp = self.api_post('/v2/images/%s/actions/deactivate' % image_id)
        self.assertEqual(204, resp.status_code)

        # Now disable reactivate permissions, but allow get_image
        self.set_policy_rules({'get_image': '',
                               'reactivate': '!'})

        # Make sure reactivate returns 403 because we can see the image,
        # just not reactivate it
        resp = self.api_post('/v2/images/%s/actions/reactivate' % image_id)
        self.assertEqual(403, resp.status_code)

        # Now disable reactivate permissions, including get_image
        self.set_policy_rules({'get_image': '!',
                               'reactivate': '!'})

        # Make sure reactivate returns 404 because we can not see nor
        # reactivate it
        resp = self.api_post('/v2/images/%s/actions/reactivate' % image_id)
        self.assertEqual(404, resp.status_code)

        # Now allow reactivate, but disallow get_image, just to prove that
        # you do not need get_image in order to be granted reactivate, and
        # that we only use it for error code determination if
        # permission is denied.
        self.set_policy_rules({'get_image': '!',
                               'reactivate': ''})

        # Make sure reactivate returns 204 because even though we can not
        # see the image, we can reactivate it
        resp = self.api_post('/v2/images/%s/actions/reactivate' % image_id)
        self.assertEqual(204, resp.status_code)

        # Make sure you can not reactivate image using non-admin role of
        # different project
        self.set_policy_rules({
            'get_image': '',
            'modify_image': '',
            'add_image': '',
            'upload_image': '',
            'add_member': '',
            'deactivate': '',
            'reactivate': '',
            'publicize_image': '',
            'communitize_image': ''
        })
        headers = self._headers({
            'X-Project-Id': 'fake-project-id',
            'X-Roles': 'member'
        })
        for visibility in ('public', 'community', 'shared', 'private'):
            image_id = self._create_and_upload(visibility=visibility)
            # deactivate the image
            resp = self.api_post(
                '/v2/images/%s/actions/deactivate' % image_id)
            self.assertEqual(204, resp.status_code)

            # try to reactivate the image
            resp = self.api_post(
                '/v2/images/%s/actions/reactivate' % image_id, headers=headers)

            # 'shared' image will return 404 until it is not shared with
            # project accessing it
            if visibility == 'shared':
                self.assertEqual(404, resp.status_code)
                # Now lets share the image and try to reactivate it
                share_path = '/v2/images/%s/members' % image_id
                data = {
                    'member': 'fake-project-id'
                }
                response = self.api_post(share_path, json=data)
                member = response.json
                self.assertEqual(200, response.status_code)
                self.assertEqual(image_id, member['image_id'])

                # Now ensure reactivating image by another tenant will
                # return 403
                resp = self.api_post(
                    '/v2/images/%s/actions/reactivate' % image_id,
                    headers=headers)
                self.assertEqual(403, resp.status_code)
            elif visibility == 'private':
                # private image will also return 404 as it is not visible
                self.assertEqual(404, resp.status_code)
            else:
                # public and community visibility will return 403
                self.assertEqual(403, resp.status_code)

    def test_delete_from_store(self):
        self.start_server()
        # First create image in multiple stores
        image_id = self._create_and_import(stores=['store1', 'store2',
                                                   'store3'])

        # Make sure we are able to delete image from the specific store
        path = "/v2/stores/store1/%s" % image_id
        response = self.api_delete(path)
        self.assertEqual(204, response.status_code)

        # Disable get_image_location and verify you will get 403
        self.set_policy_rules({
            'get_image': '',
            'delete_image_location': '',
            'get_image_location': '!'
        })
        path = "/v2/stores/store2/%s" % image_id
        response = self.api_delete(path)
        self.assertEqual(403, response.status_code)

        # Disable delete_image_location and verify you will get 403
        self.set_policy_rules({
            'get_image': '',
            'delete_image_location': '!',
            'get_image_location': ''
        })
        path = "/v2/stores/store2/%s" % image_id
        response = self.api_delete(path)
        self.assertEqual(403, response.status_code)

        # Disabling all, you will get 404
        self.set_policy_rules({
            'get_image': '!',
            'delete_image_location': '!',
            'get_image_location': '!'
        })
        path = "/v2/stores/store2/%s" % image_id
        response = self.api_delete(path)
        self.assertEqual(404, response.status_code)

        # Now allow delete_image_location and get_image_location, but disallow
        # get_image, just to prove that you do not need get_image in order
        # to be granted delete image from particular store, and
        # that we only use it for error code determination if
        # permission is denied.
        self.set_policy_rules({
            'get_image': '!',
            'delete_image_location': '',
            'get_image_location': ''
        })
        path = "/v2/stores/store2/%s" % image_id
        response = self.api_delete(path)
        self.assertEqual(204, response.status_code)

        # deleting image with non-admin will get 403
        self.set_policy_rules({
            'get_image': '',
            'delete_image_location': '',
            'get_image_location': ''
        })
        headers = self._headers({
            'X-Roles': 'member'
        })

        path = "/v2/stores/store2/%s" % image_id
        response = self.api_delete(path, headers=headers)
        self.assertEqual(403, response.status_code)

    def test_copy_image(self):
        self.start_server()
        # create image using import
        image_id = self._create_and_import(
            stores=['store1'], visibility='public')

        # Make sure you can copy image to another store
        self.set_policy_rules({
            'copy_image': 'role:admin',
            'get_image': '',
            'modify_image': ''
        })
        store_to_copy = ["store2"]
        response = self._import_copy(image_id, store_to_copy)
        self.assertEqual(202, response.status_code)
        self._wait_for_import(image_id)
        self.assertEqual('success', self._get_latest_task(image_id)['status'])

        # Now disable copy image and see you will get 403 Forbidden
        store_to_copy = ["store3"]
        self.set_policy_rules({
            'copy_image': '!',
            'get_image': '',
            'modify_image': ''
        })
        response = self._import_copy(image_id, store_to_copy)
        self.assertEqual(403, response.status_code)

        # Verify that non-admin but member of same project can not copy image
        self.set_policy_rules({
            'copy_image': 'role:admin',
            'get_image': '',
            'modify_image': ''
        })
        headers = self._headers({'X-Roles': 'member'})
        response = self._import_copy(image_id, store_to_copy,
                                     headers=headers)
        self.assertEqual(403, response.status_code)

        # Verify that non-owner can not copy image
        self.set_policy_rules({
            'copy_image': 'role:admin',
            'get_image': '',
            'modify_image': ''
        })
        headers = self._headers({
            'X-Roles': 'member',
            'X-Project-Id': 'fake-project-id'
        })
        response = self._import_copy(image_id, store_to_copy,
                                     headers=headers)
        self.assertEqual(403, response.status_code)

        # Now disable copy image and get_image and see you will get
        # 404 NotFound
        self.set_policy_rules({
            'copy_image': '!',
            'get_image': '!',
            'modify_image': ''
        })
        store_to_copy = ["store3"]
        print(self.policy.rules.items())
        response = self._import_copy(image_id, store_to_copy)
        self.assertEqual(404, response.status_code)

    def test_import_glance_direct(self):
        self.start_server()
        # create image and stage data to it
        image_id = self._create_and_stage(visibility='public')

        # Make sure you can import using glance-direct
        self.set_policy_rules({
            'get_image': '',
            'communitize_image': '',
            'add_image': '',
            'modify_image': ''
        })
        store_to_import = ['store1']
        response = self._import_direct(image_id, store_to_import)
        self.assertEqual(202, response.status_code)
        self._wait_for_import(image_id)
        self.assertEqual('success', self._get_latest_task(image_id)['status'])

        # Make sure you can import data to image using non-admin role
        image_id = self._create_and_stage(visibility='community')
        headers = self._headers({'X-Roles': 'member'})
        response = self._import_direct(image_id, store_to_import,
                                       headers=headers)
        self.assertEqual(202, response.status_code)
        self._wait_for_import(image_id)
        self.assertEqual('success', self._get_latest_task(image_id)['status'])

        # Make sure you can not import data to image using non-admin role of
        # different project
        image_id = self._create_and_stage(visibility='community')
        # Make sure you will get 403 Forbidden
        self.set_policy_rules({
            'get_image': '',
            'modify_image': '!'
        })
        headers = self._headers({
            'X-Roles': 'member',
            'X-Project-Id': 'fake-project-id'
        })
        response = self._import_direct(image_id, store_to_import,
                                       headers=headers)
        self.assertEqual(403, response.status_code)

        # disabling both get_image and modify_image should return 404 NotFound
        self.set_policy_rules({
            'get_image': '!',
            'modify_image': '!'
        })
        headers = self._headers({
            'X-Roles': 'member',
            'X-Project-Id': 'fake-project-id'
        })
        response = self._import_direct(image_id, store_to_import,
                                       headers=headers)
        self.assertEqual(404, response.status_code)

    def _test_image_ownership(self, headers, method):
        self.set_policy_rules({
            'get_image': '',
            'add_image': '',
            'publicize_image': '',
            'communitize_image': '',
            'add_member': '',
        })

        for visibility in ('community', 'public', 'shared'):
            path = "/v2/images"
            data = {
                'name': '%s-image' % visibility,
                'visibility': visibility,
            }
            # create image
            response = self.api_post(path, json=data)
            image = response.json
            self.assertEqual(201, response.status_code)
            self.assertEqual(visibility, image['visibility'])

            # share the image if visibility is shared
            if visibility == 'shared':
                path = '/v2/images/%s/members' % image['id']
                data = {
                    'member': 'fake-project-id'
                }
                response = self.api_post(path,
                                         json=data)
                self.assertEqual(200, response.status_code)

            # Add/Delete tag
            path = '/v2/images/%s/tags/Test_Tag_2' % image['id']
            response = self.api_request(method, path, headers=headers)
            self.assertEqual(403, response.status_code)

    def test_image_tag_update(self):
        self.start_server()
        # Create image
        image_id = self._create_and_upload()

        # Make sure we will be able to add tags for the image
        path = '/v2/images/%s/tags/Test_Tag' % image_id
        response = self.api_put(path)
        self.assertEqual(204, response.status_code)
        # Make sure tag is added to image
        path = '/v2/images/%s' % image_id
        response = self.api_get(path)
        image = response.json
        self.assertEqual(['Test_Tag'], image['tags'])

        # Disable get_image and modify_image should give us 404 Not Found
        self.set_policy_rules({
            'get_image': '!',
            'modify_image': '!'
        })
        path = '/v2/images/%s/tags/Test_Tag_2' % image_id
        response = self.api_put(path)
        self.assertEqual(404, response.status_code)

        # Allow get_image and disable modify_image should give us
        # 403 Forbidden
        self.set_policy_rules({
            'get_image': '',
            'modify_image': '!'
        })
        path = '/v2/images/%s/tags/Test_Tag_2' % image_id
        response = self.api_put(path)
        self.assertEqual(403, response.status_code)

        # Adding tag by another project (non-admin user) should return
        # 404 Not Found for private image
        self.set_policy_rules({
            'get_image': '',
            'modify_image': ''
        })
        # Note for other reviewers, these tests runs by default using
        # admin role, to test this scenario we need image
        # of current project to be accessed by other projects non-admin
        # user.
        headers = self._headers({
            'X-Project-Id': 'fake-project-id',
            'X-Roles': 'member',
        })
        path = '/v2/images/%s/tags/Test_Tag_2' % image_id
        response = self.api_put(path, headers=headers)
        self.assertEqual(404, response.status_code)

        # Adding tag by another project (non-admin user) should return
        # 403 Not Found for other than private image
        self._test_image_ownership(headers, 'PUT')

    def test_image_tag_delete(self):
        self.start_server()
        # Create image
        image_id = self._create_and_upload()

        # Make sure we will be able to add tags for the image
        path = '/v2/images/%s/tags/Test_Tag_1' % image_id
        response = self.api_put(path)
        self.assertEqual(204, response.status_code)
        # add another tag while we can
        path = '/v2/images/%s/tags/Test_Tag_2' % image_id
        response = self.api_put(path)
        self.assertEqual(204, response.status_code)

        # Make sure tags are added to image
        path = '/v2/images/%s' % image_id
        response = self.api_get(path)
        image = response.json
        self.assertItemsEqual(['Test_Tag_1', 'Test_Tag_2'], image['tags'])

        # Now delete tag from image
        path = '/v2/images/%s/tags/Test_Tag_1' % image_id
        response = self.api_delete(path)
        self.assertEqual(204, response.status_code)

        # Make sure tag is deleted
        path = '/v2/images/%s' % image_id
        response = self.api_get(path)
        image = response.json
        self.assertNotIn('Test_Tag_1', image['tags'])

        # Disable get_image and modify_image should give us 404 Not Found
        self.set_policy_rules({
            'get_image': '!',
            'modify_image': '!'
        })
        path = '/v2/images/%s/tags/Test_Tag_2' % image_id
        response = self.api_delete(path)
        self.assertEqual(404, response.status_code)

        # Allow get_image and disable modify_image should give us
        # 403 Forbidden
        self.set_policy_rules({
            'get_image': '',
            'modify_image': '!'
        })
        path = '/v2/images/%s/tags/Test_Tag_2' % image_id
        response = self.api_delete(path)
        self.assertEqual(403, response.status_code)

        # Deleting tag by another project (non-admin user) should return
        # 404 Not Found for private image
        self.set_policy_rules({
            'get_image': '',
            'modify_image': ''
        })
        # Note for other reviewers, these tests runs by default using
        # admin role, to test this scenario we need image
        # of current project to be accessed by other projects non-admin
        # user.
        headers = self._headers({
            'X-Project-Id': 'fake-project-id',
            'X-Roles': 'member',
        })
        path = '/v2/images/%s/tags/Test_Tag_2' % image_id
        response = self.api_delete(path, headers=headers)
        self.assertEqual(404, response.status_code)

        # Deleting tag by another project (non-admin user) should return
        # 403 Not Found for other than private image
        self._test_image_ownership(headers, 'DELETE')

    def test_get_task_info(self):
        self.start_server()
        image_id = self._create_and_import(
            stores=['store1'], visibility='public')

        # Make sure you can get task information of that image
        path = '/v2/images/%s/tasks' % image_id
        response = self.api_get(path)
        self.assertEqual(200, response.status_code)

        # Disable get_image should give us 404 Not Found
        self.set_policy_rules({'get_image': '!'})
        path = '/v2/images/%s/tasks' % image_id
        response = self.api_get(path)
        self.assertEqual(404, response.status_code)



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from stevedore import named

from glance.i18n import _


CONF = cfg.CONF

import_filtering_opts = [

    cfg.ListOpt('allowed_schemes',
                item_type=cfg.types.String(quotes=True),
                bounds=True,
                default=['http', 'https'],
                help=_("""
Specify the "whitelist" of allowed url schemes for web-download.

This option provides whitelisting of uri schemes that will be allowed when
an end user imports an image using the web-download import method. The
whitelist has priority such that if there is also a blacklist defined for
schemes, the blacklist will be ignored.  Host and port filtering, however,
will be applied.

See the Glance Administration Guide for more information.

Possible values:
    * List containing normalized url schemes as they are returned from
      urllib.parse. For example ['ftp','https']
    * Hint: leave the whitelist empty if you want the disallowed_schemes
      blacklist to be processed

Related options:
    * disallowed_schemes
    * allowed_hosts
    * disallowed_hosts
    * allowed_ports
    * disallowed_ports

""")),
    cfg.ListOpt('disallowed_schemes',
                item_type=cfg.types.String(quotes=True),
                bounds=True,
                default=[],
                help=_("""
Specify the "blacklist" of uri schemes disallowed for web-download.

This option provides blacklisting of uri schemes that will be rejected when
an end user imports an image using the web-download import method.  Note
that if a scheme whitelist is defined using the 'allowed_schemes' option,
*this option will be ignored*.  Host and port filtering, however, will be
applied.

See the Glance Administration Guide for more information.

Possible values:
    * List containing normalized url schemes as they are returned from
      urllib.parse. For example ['ftp','https']
    * By default the list is empty

Related options:
    * allowed_schemes
    * allowed_hosts
    * disallowed_hosts
    * allowed_ports
    * disallowed_ports

""")),
    cfg.ListOpt('allowed_hosts',
                item_type=cfg.types.HostAddress(),
                bounds=True,
                default=[],
                help=_("""
Specify the "whitelist" of allowed target hosts for web-download.

This option provides whitelisting of hosts that will be allowed when an end
user imports an image using the web-download import method. The whitelist
has priority such that if there is also a blacklist defined for hosts, the
blacklist will be ignored.  The uri must have already passed scheme
filtering before this host filter will be applied.  If the uri passes, port
filtering will then be applied.

See the Glance Administration Guide for more information.

Possible values:
    * List containing normalized hostname or ip like it would be returned
      in the urllib.parse netloc without the port
    * By default the list is empty
    * Hint: leave the whitelist empty if you want the disallowed_hosts
      blacklist to be processed

Related options:
    * allowed_schemes
    * disallowed_schemes
    * disallowed_hosts
    * allowed_ports
    * disallowed_ports

""")),
    cfg.ListOpt('disallowed_hosts',
                item_type=cfg.types.HostAddress(),
                bounds=True,
                default=[],
                help=_("""
Specify the "blacklist" of hosts disallowed for web-download.

This option provides blacklisting of hosts that will be rejected when an end
user imports an image using the web-download import method.  Note that if a
host whitelist is defined using the 'allowed_hosts' option, *this option
will be ignored*.

The uri must have already passed scheme filtering before this host filter
will be applied.  If the uri passes, port filtering will then be applied.

See the Glance Administration Guide for more information.

Possible values:
    * List containing normalized hostname or ip like it would be returned
      in the urllib.parse netloc without the port
    * By default the list is empty

Related options:
    * allowed_schemes
    * disallowed_schemes
    * allowed_hosts
    * allowed_ports
    * disallowed_ports

""")),
    cfg.ListOpt('allowed_ports',
                item_type=cfg.types.Integer(min=1, max=65535),
                bounds=True,
                default=[80, 443],
                help=_("""
Specify the "whitelist" of allowed ports for web-download.

This option provides whitelisting of ports that will be allowed when an end
user imports an image using the web-download import method.  The whitelist
has priority such that if there is also a blacklist defined for ports, the
blacklist will be ignored.  Note that scheme and host filtering have already
been applied by the time a uri hits the port filter.

See the Glance Administration Guide for more information.

Possible values:
    * List containing ports as they are returned from urllib.parse netloc
      field.  Thus the value is a list of integer values, for example
      [80, 443]
    * Hint: leave the whitelist empty if you want the disallowed_ports
      blacklist to be processed

Related options:
    * allowed_schemes
    * disallowed_schemes
    * allowed_hosts
    * disallowed_hosts
    * disallowed_ports
""")),
    cfg.ListOpt('disallowed_ports',
                item_type=cfg.types.Integer(min=1, max=65535),
                bounds=True,
                default=[],
                help=_("""
Specify the "blacklist" of disallowed ports for web-download.

This option provides blacklisting of target ports that will be rejected when
an end user imports an image using the web-download import method.  Note
that if a port whitelist is defined using the 'allowed_ports' option, *this
option will be ignored*.  Note that scheme and host filtering have already
been applied by the time a uri hits the port filter.

See the Glance Administration Guide for more information.

Possible values:
    * List containing ports as they are returned from urllib.parse netloc
      field.  Thus the value is a list of integer values, for example
      [22, 88]
    * By default this list is empty

Related options:
    * allowed_schemes
    * disallowed_schemes
    * allowed_hosts
    * disallowed_hosts
    * allowed_ports

""")),
]

CONF.register_opts(import_filtering_opts, group='import_filtering_opts')


def get_import_plugin(**kwargs):
    method_list = CONF.enabled_import_methods
    import_method = kwargs.get('import_req')['method']['name']
    if import_method in method_list:
        import_method = import_method.replace("-", "_")
        task_list = [import_method]
        # TODO(jokke): Implement error handling of non-listed methods.
    extensions = named.NamedExtensionManager(
        'glance.image_import.internal_plugins',
        names=task_list,
        name_order=True,
        invoke_on_load=True,
        invoke_kwds=kwargs)
    for extension in extensions.extensions:
        return extension.obj



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
import urllib
import urllib.error
import urllib.request

from glance.common import exception
from glance.common.scripts import utils as script_utils
import glance.tests.utils as test_utils


class TestScriptsUtils(test_utils.BaseTestCase):
    def setUp(self):
        super(TestScriptsUtils, self).setUp()

    def test_get_task(self):
        task = mock.ANY
        task_repo = mock.Mock(return_value=task)
        task_id = mock.ANY
        self.assertEqual(task, script_utils.get_task(task_repo, task_id))

    def test_unpack_task_input(self):
        task_input = {"import_from": "foo",
                      "import_from_format": "bar",
                      "image_properties": "baz"}
        task = mock.Mock(task_input=task_input)
        self.assertEqual(task_input,
                         script_utils.unpack_task_input(task))

    def test_unpack_task_type_location_import(self):
        task_type = 'location_import'
        task_input = {'image_id': mock.ANY,
                      'loc_url': mock.ANY,
                      'validation_data': {}}
        task = mock.Mock(type=task_type, task_input=task_input)
        self.assertEqual(task_input,
                         script_utils.unpack_task_input(task))

    def test_unpack_task_type_location_import_error(self):
        task_type = 'location_import'
        task_input1 = {'image_id': mock.ANY,
                       'validation_data': {}}
        task_input2 = {'loc_url': mock.ANY,
                       'validation_data': {}}
        task_input3 = {'image_id': mock.ANY,
                       'loc_url': mock.ANY}
        task1 = mock.Mock(type=task_type, task_input=task_input1)
        task2 = mock.Mock(type=task_type, task_input=task_input2)
        task3 = mock.Mock(type=task_type, task_input=task_input3)
        self.assertRaises(exception.Invalid,
                          script_utils.unpack_task_input, task1)
        self.assertRaises(exception.Invalid,
                          script_utils.unpack_task_input, task2)
        self.assertRaises(exception.Invalid,
                          script_utils.unpack_task_input, task3)

    def test_unpack_task_input_error(self):
        task_input1 = {"import_from_format": "bar", "image_properties": "baz"}
        task_input2 = {"import_from": "foo", "image_properties": "baz"}
        task_input3 = {"import_from": "foo", "import_from_format": "bar"}
        task1 = mock.Mock(task_input=task_input1)
        task2 = mock.Mock(task_input=task_input2)
        task3 = mock.Mock(task_input=task_input3)
        self.assertRaises(exception.Invalid,
                          script_utils.unpack_task_input, task1)
        self.assertRaises(exception.Invalid,
                          script_utils.unpack_task_input, task2)
        self.assertRaises(exception.Invalid,
                          script_utils.unpack_task_input, task3)

    def test_set_base_image_properties(self):
        properties = {}
        script_utils.set_base_image_properties(properties)
        self.assertIn('disk_format', properties)
        self.assertIn('container_format', properties)
        self.assertEqual('qcow2', properties['disk_format'])
        self.assertEqual('bare', properties['container_format'])

    def test_set_base_image_properties_none(self):
        properties = None
        script_utils.set_base_image_properties(properties)
        self.assertIsNone(properties)

    def test_set_base_image_properties_not_empty(self):
        properties = {'disk_format': 'vmdk', 'container_format': 'bare'}
        script_utils.set_base_image_properties(properties)
        self.assertIn('disk_format', properties)
        self.assertIn('container_format', properties)
        self.assertEqual('vmdk', properties.get('disk_format'))
        self.assertEqual('bare', properties.get('container_format'))

    def test_validate_location_http(self):
        location = 'http://example.com'
        self.assertEqual(location,
                         script_utils.validate_location_uri(location))

    def test_validate_location_https(self):
        location = 'https://example.com'
        self.assertEqual(location,
                         script_utils.validate_location_uri(location))

    def test_validate_location_none_error(self):
        self.assertRaises(exception.BadStoreUri,
                          script_utils.validate_location_uri, '')

    def test_validate_location_file_location_error(self):
        self.assertRaises(exception.BadStoreUri,
                          script_utils.validate_location_uri, "file:///tmp")
        self.assertRaises(exception.BadStoreUri,
                          script_utils.validate_location_uri,
                          "filesystem:///tmp")

    def test_validate_location_unsupported_error(self):
        location = 'swift'
        self.assertRaises(urllib.error.URLError,
                          script_utils.validate_location_uri, location)

        location = 'swift+http'
        self.assertRaises(urllib.error.URLError,
                          script_utils.validate_location_uri, location)

        location = 'swift+https'
        self.assertRaises(urllib.error.URLError,
                          script_utils.validate_location_uri, location)

        location = 'swift+config'
        self.assertRaises(urllib.error.URLError,
                          script_utils.validate_location_uri, location)

        location = 'vsphere'
        self.assertRaises(urllib.error.URLError,
                          script_utils.validate_location_uri, location)

        location = 'rbd://'
        self.assertRaises(urllib.error.URLError,
                          script_utils.validate_location_uri, location)

        location = 'cinder://'
        self.assertRaises(urllib.error.URLError,
                          script_utils.validate_location_uri, location)


class TestCallbackIterator(test_utils.BaseTestCase):
    def test_iterator_iterates(self):
        # Include a zero-length generation to make sure we don't trigger
        # the callback when nothing new has happened.
        items = ['1', '2', '', '3']
        callback = mock.MagicMock()
        cb_iter = script_utils.CallbackIterator(iter(items), callback)
        iter_items = list(cb_iter)
        callback.assert_has_calls([mock.call(1, 1),
                                   mock.call(1, 2),
                                   mock.call(1, 3)])
        self.assertEqual(items, iter_items)

        # Make sure we don't call the callback on close if we
        # have processed all the data
        callback.reset_mock()
        cb_iter.close()
        callback.assert_not_called()

    @mock.patch('oslo_utils.timeutils.StopWatch')
    def test_iterator_iterates_granularly(self, mock_sw):
        items = ['1', '2', '3']
        callback = mock.MagicMock()
        mock_sw.return_value.expired.side_effect = [False, True, False]
        cb_iter = script_utils.CallbackIterator(iter(items), callback,
                                                min_interval=30)
        iter_items = list(cb_iter)
        self.assertEqual(items, iter_items)
        # The timer only fired once, but we should still expect the final
        # chunk to be emitted.
        callback.assert_has_calls([mock.call(2, 2),
                                   mock.call(1, 3)])

        mock_sw.assert_called_once_with(30)
        mock_sw.return_value.start.assert_called_once_with()
        mock_sw.return_value.restart.assert_called_once_with()

        # Make sure we don't call the callback on close if we
        # have processed all the data
        callback.reset_mock()
        cb_iter.close()
        callback.assert_not_called()

    def test_proxy_close(self):
        callback = mock.MagicMock()
        source = mock.MagicMock()
        del source.close
        # NOTE(danms): This will generate AttributeError if it
        # tries to call close after the del above.
        script_utils.CallbackIterator(source, callback).close()

        source = mock.MagicMock()
        source.close.return_value = 'foo'
        script_utils.CallbackIterator(source, callback).close()
        source.close.assert_called_once_with()

        # We didn't process any data, so no callback should be expected
        callback.assert_not_called()

    @mock.patch('oslo_utils.timeutils.StopWatch')
    def test_proxy_read(self, mock_sw):
        items = ['1', '2', '3']
        source = mock.MagicMock()
        source.read.side_effect = items
        callback = mock.MagicMock()
        mock_sw.return_value.expired.side_effect = [False, True, False]
        cb_iter = script_utils.CallbackIterator(source, callback,
                                                min_interval=30)
        results = [cb_iter.read(1) for i in range(len(items))]
        self.assertEqual(items, results)
        # The timer only fired once while reading, so we only expect
        # one callback.
        callback.assert_has_calls([mock.call(2, 2)])
        cb_iter.close()
        # If we close with residue since the last callback, we should
        # call the callback with that.
        callback.assert_has_calls([mock.call(2, 2),
                                   mock.call(1, 3)])


class TestSafeRedirectHandler(test_utils.BaseTestCase):
    """Test SafeRedirectHandler for redirect validation."""

    def setUp(self):
        super(TestSafeRedirectHandler, self).setUp()

    @mock.patch('glance.common.utils.validate_import_uri')
    def test_redirect_to_allowed_url(self, mock_validate):
        """Test redirect to allowed URL is accepted."""
        mock_validate.return_value = True
        handler = script_utils.SafeRedirectHandler()

        req = mock.Mock()
        req.full_url = 'http://example.com/redirect'
        fp = mock.Mock()
        headers = mock.Mock()

        # Redirect to allowed URL
        # redirect_request should call super().redirect_request
        # which returns a request
        with mock.patch.object(urllib.request.HTTPRedirectHandler,
                               'redirect_request') as mock_super:
            mock_super.return_value = mock.Mock()
            result = handler.redirect_request(
                req, fp, 302, 'Found', headers, 'http://allowed.com/target'
            )

        mock_validate.assert_called_once_with('http://allowed.com/target')
        # Should return a request object (not None)
        self.assertIsNotNone(result)

    @mock.patch('glance.common.utils.validate_import_uri')
    def test_redirect_to_disallowed_url(self, mock_validate):
        """Test redirect to disallowed URL raises error."""
        mock_validate.return_value = False
        handler = script_utils.SafeRedirectHandler()

        req = mock.Mock()
        req.full_url = 'http://example.com/redirect'
        fp = mock.Mock()
        headers = mock.Mock()

        # Redirect to disallowed URL should raise ImportTaskError
        self.assertRaises(
            exception.ImportTaskError,
            handler.redirect_request,
            req, fp, 302, 'Found', headers, 'http://127.0.0.1:5000/'
        )

        mock_validate.assert_called_once_with('http://127.0.0.1:5000/')


class TestGetImageDataIter(test_utils.BaseTestCase):
    """Test get_image_data_iter with redirect validation."""

    def setUp(self):
        super(TestGetImageDataIter, self).setUp()

    @mock.patch('builtins.open', create=True)
    def test_get_image_data_iter_file_uri(self, mock_open):
        """Test file:// URI handling."""
        mock_file = mock.Mock()
        mock_open.return_value = mock_file

        result = script_utils.get_image_data_iter("file:///tmp/test.img")

        mock_open.assert_called_once_with("/tmp/test.img", "rb")
        self.assertEqual(result, mock_file)

    @mock.patch('urllib.request.build_opener')
    def test_get_image_data_iter_http_uri(self, mock_build_opener):
        """Test HTTP URI handling with redirect validation."""
        mock_opener = mock.Mock()
        mock_response = mock.Mock()
        mock_opener.open.return_value = mock_response
        mock_build_opener.return_value = mock_opener

        result = script_utils.get_image_data_iter("http://example.com/image")

        # Should use build_opener with SafeRedirectHandler
        mock_build_opener.assert_called_once()
        # Check that SafeRedirectHandler was passed as an argument
        call_args = mock_build_opener.call_args
        # build_opener can be called with *args or keyword args
        # Check both positional and keyword arguments
        found_handler = False
        if call_args.args:
            found_handler = any(
                isinstance(arg, script_utils.SafeRedirectHandler)
                for arg in call_args.args)
        if not found_handler and call_args.kwargs:
            found_handler = any(
                isinstance(v, script_utils.SafeRedirectHandler)
                for v in call_args.kwargs.values())
        # Also check if it's passed as a handler class (not instance)
        if not found_handler:
            found_handler = (
                script_utils.SafeRedirectHandler in call_args.args)

        self.assertTrue(
            found_handler,
            "SafeRedirectHandler should be passed to build_opener")

        mock_opener.open.assert_called_once_with("http://example.com/image")
        self.assertEqual(result, mock_response)
